import unittest
import numpy as np
import pypolar.jones as jones
import pypolar.mueller as mueller

class TestBasic(unittest.TestCase):

    def test_field_linear_H(self):
        H = mueller.stokes_horizontal()
        S = mueller.stokes_linear(0)
        for pair in zip(H, S):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_field_linear_V(self):
        V = mueller.stokes_vertical()
        S = mueller.stokes_linear(np.pi/2)
        for pair in zip(V, S):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_field_linear_multi(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        H = mueller.stokes_horizontal()
        V = mueller.stokes_vertical()
        S = mueller.stokes_linear(angles)
        self.assertEqual(len(S),N)

    def test_intensity_scalar(self):
        S = mueller.stokes_right_circular()
        I = mueller.intensity(S)
        self.assertAlmostEqual(I,1)
        S = mueller.stokes_horizontal()
        I = mueller.intensity(S)
        self.assertAlmostEqual(I,1)

    def test_intensity_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        S = mueller.stokes_linear(angles)
        I = mueller.intensity(S)
        self.assertEqual(len(I),N)
        for intensity in I:
            self.assertAlmostEqual(intensity, 1)

    def test_dop_scalar(self):
        S = mueller.stokes_left_circular()
        dop = mueller.degree_of_polarization(S)
        self.assertAlmostEqual(dop, 1)
        S = mueller.stokes_unpolarized()
        print(S)
        dop = mueller.degree_of_polarization(S)
        self.assertAlmostEqual(dop, 0)
        S = mueller.stokes_elliptical(0.5, np.pi/6, np.pi/3)
        dop = mueller.degree_of_polarization(S)
        self.assertAlmostEqual(dop, 0.5)

    def test_dop_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        S = mueller.stokes_linear(angles)
        dop = mueller.degree_of_polarization(S)
        self.assertEqual(len(dop),N)
        for p in dop:
            self.assertAlmostEqual(p, 1)

    def test_to_jones_scalar(self):
        S = mueller.stokes_left_circular()
        J = jones.field_left_circular()
        JJ = mueller.stokes_to_jones(S)
        for pair in zip(J, JJ):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_to_stokes_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        S = mueller.stokes_linear(angles)
        J = mueller.stokes_to_jones(S)
        n,m = J.shape
        self.assertEqual(n,N)
        self.assertEqual(m,2)

if __name__ == '__main__':
    unittest.main()


# pylint: disable=invalid-name
# pylint: disable=bare-except
# pylint: disable=no-member

"""
Useful routines for symbolic manipulation of Jones vectors and matrices.

Creating Jones vectors for specific polarization states::

    * field_linear(angle)
    * field_left_circular()
    * field_right_circular()
    * field_horizontal()
    * field_vertical()
    * field_ellipsometry(tanpsi, Delta)
    * field_elliptical(azimuth, elliptic_angle)

Creating Jones Matrices for polarizing elements::

    * op_linear_polarizer(angle)
    * op_retarder(fast_axis_angle, retardance)
    * op_attenuator(optical_density)
    * op_mirror()
    * op_rotation(angle)
    * op_quarter_wave_plate(fast_axis_angle)
    * op_half_wave_plate(fast_axis_angle)
    * op_fresnel_reflection(index_of_refraction, angle)
    * op_fresnel_transmission(index_of_refraction, angle)

Interpreting the polarization state::

    * use_alternate_convention(boolean)
    * interpret(jones_vector)
    * intensity(jones_vector)
    * phase(jones_vector)
    * ellipse_azimuth(jones_vector)
    * ellipse_axes(jones_vector)
    * ellipticity(jones_vector)
    * ellipticity_angle(jones_vector)
    * amplitude_ratio(jones_vector)
    * amplitude_ratio_angle(jones_vector)
    * polarization_variable(jones_vector)

Converting to Mueller formalism::

    * jones_op_to_mueller_op(jones_matrix)
    * jones_to_stokes(jones_vector)
"""

import sympy
import pypolar.sym_fresnel as sym_fresnel

__all__ = ('op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'field_linear',
           'field_left_circular',
           'field_right_circular',
           'field_horizontal',
           'field_vertical',
           'intensity',
           'phase')


def op_linear_polarizer(theta):
    """
    Jones matrix operator for a rotated linear polarizer.

    The polarizer is rotated around a normal to its surface.

    Args:
        theta: rotation angle measured from the horizontal plane [radians]
    """
    return sympy.Matrix([[sympy.cos(theta)**2, sympy.sin(theta) * sympy.cos(theta)],
                         [sympy.sin(theta) * sympy.cos(theta), sympy.sin(theta)**2]])


def op_retarder(theta, delta):
    """
    Jones matrix operator for a rotated optical retarder.

    The retarder is rotated around a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
        delta: phase delay introduced between fast and slow-axes         [radians]
    """
    P = sympy.exp(+delta / 2 * sympy.I)
    Q = sympy.exp(-delta / 2 * sympy.I)
    D = sympy.sin(delta / 2) * 2 * sympy.I
    C = sympy.cos(theta)
    S = sympy.sin(theta)
    return sympy.Matrix([[C * C * P + S * S * Q, C * S * D],
                         [C * S * D, C * C * Q + S * S * P]])


def op_attenuator(t):
    """
    Jones matrix operator for an optical attenuator.

    Args:
        t: fraction of intensity passing through attenuator [---]
    """
    f = sympy.sqrt(t)
    return sympy.Matrix([[f, 0], [0, f]])


def op_neutral_density_filter(nd):
    """
    Jones matrix operator for a neutral density filter with decadic attenuation.

    Args:
        nd: base ten optical density  [---]
    """
    return sympy.Matrix([[10**-nd / 2, 0], [0, 10**-nd / 2]])

def op_mirror():
    """Jones matrix operator for a perfect mirror."""
    return sympy.Matrix([[1, 0], [0, -1]])


def op_rotation(theta):
    """
    Jones matrix operator to rotate light around the optical axis.

    Args:
        theta : angle of rotation about optical axis  [radians]
    Returns:
        2x2 matrix of the rotation operator           [-]
    """
    return sympy.Matrix([[sympy.cos(theta), sympy.sin(theta)],
                         [-sympy.sin(theta), sympy.cos(theta)]])


def op_quarter_wave_plate(theta):
    """
    Jones matrix operator for a rotated quarter-wave plate.

    The QWP is rotated about a normal to its surface.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the quarter-wave plate operator     [-]
    """
    return op_retarder(theta, sympy.pi / 2)


def op_half_wave_plate(theta):
    """
    Jones matrix operator for a rotated half-wave plate.

    The HWP is rotated about a normal to its surface.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the half-wave plate operator     [-]
    """
    return op_retarder(theta, sympy.pi)


def op_fresnel_reflection(m, theta):
    """
    Jones matrix operator for Fresnel reflection at angle.

    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        2x2 matrix of the Fresnel transmission operator     [-]
    """
    return sympy.Matrix([[sym_fresnel.r_par_amplitude(m, theta), 0],
                         [0, sym_fresnel.r_per_amplitude(m, theta)]])


def op_fresnel_transmission(m, theta):
    """
    Jones matrix operator for Fresnel transmission at angle theta.

    *** THIS IS ALMOST CERTAINLY WRONG ***

    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        2x2 Fresnel transmission operator           [-]
    """
    c = sympy.cos(theta)
    d = sympy.sqrt(m * m - sympy.sin(theta)**2)
    if m.imag == 0:
        d = sympy.conjugate(d)
    a = sympy.sqrt(d/c)
    tpar = sym_fresnel.t_par_amplitude(m, theta)
    tper = sym_fresnel.t_per_amplitude(m, theta)
    return a*sympy.Matrix([[tpar, 0], [0, tper]])


def field_linear(theta):
    """Jones vector for linear polarized light at angle theta from horizontal plane."""
    return sympy.Matrix([sympy.cos(theta), sympy.sin(theta)])


def field_right_circular():
    """Jones Vector for right circular polarized light."""
    return 1 / sympy.sqrt(2) * sympy.Matrix([1, -sympy.I])


def field_left_circular():
    """Jones Vector for left circular polarized light."""
    return 1 / sympy.sqrt(2) * sympy.Matrix([1, sympy.I])


def field_horizontal():
    """Jones Vector for horizontal polarized light."""
    return field_linear(0)


def field_vertical():
    """Jones Vector for vertical polarized light."""
    return field_linear(sympy.pi / 2)


def field_elliptical(A, B):
    """Jones Vector for elliptically polarized light."""
    return sympy.Matrix([A, B])


def intensity(J):
    """Return the intensity."""
    return sympy.conjugate(J.T).dot(J)


def phase(J):
    """Return the phase."""
    gamma = sympy.arg(J[1]) - sympy.arg(J[0])
    return gamma


def ellipse_orientation(J):
    """
    Return the angle between the major semi-axis and the x-axis.

    This angle is sometimes called the azimuth or psi.
    """
    Ex = sympy.Abs(J[0, :])
    Ey = sympy.Abs(J[1, :])
    delta = phase(J)
    numer = 2 * Ex * Ey * sympy.cos(delta)
    denom = Ex**2 - Ey**2
    psi = 0.5 * sympy.arctan2(numer, denom)
    return psi


def ellipse_ellipticity(J):
    """Return the ellipticity of the polarization ellipse."""
    delta = phase(J)
    psi = ellipse_orientation(J)
    chi = 0.5 * sympy.arcsin(sympy.sin(2 * psi) * sympy.sin(delta))
    return chi


def ellipse_axes(J):
    """Return the semi-major and semi-minor axes of the polarization ellipse."""
    Exo, Eyo = sympy.conjugate(J.T)*J
    psi = ellipse_orientation(J)
    delta = phase(J)
    C = sympy.cos(psi)
    S = sympy.sin(psi)
    asqr = (Exo * C)**2 + (Eyo * S)**2 + 2 * Exo * Eyo * C * S * sympy.cos(delta)
    bsqr = (Exo * S)**2 + (Eyo * C)**2 - 2 * Exo * Eyo * C * S * sympy.cos(delta)
    return sympy.sqrt(abs(asqr)), sympy.sqrt(abs(bsqr))



# pylint: disable=invalid-name
# pylint: disable=bare-except
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=unused-import
# pylint: disable=too-many-arguments

"""
A set of basic routines for visualizing polarization.

Functions for drawing the polarization ellipse (sectional pattern)::

   draw_jones_ellipse(J)
   draw_stokes_ellipse(S)

Functions for drawing 2D and 3D representations::

    draw_jones_field(J)
    draw_stokes_field(S)

Functions for drawing an animated 2D and 3D representations::

   draw_jones_animated(J)
   draw_stokes_animated(S)

Functions for drawing a Poincaré representation::
   draw_empty_sphere()
   draw_jones_poincare(J)
   draw_stokes_poincare(S)
   join_jones_poincare(J)
   join_stokes_poincare(S)

Example: Poincaré sphere plot of a Jones vector::

    J = pypolar.jones.field_linear(np.pi/6)
    pypolar.visualization.draw_jones_poincare(J)

Example: Poincaré sphere plot of two Stokes vectors::

    S1 = pypolar.mueller.stokes_left_circular()
    S2 = pypolar.mueller.stokes_linear(np.radians(15))

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection='3d')
    pypolar.visualization.draw_empty_sphere(ax)
    pypolar.visualization.draw_stokes_poincare(S1, ax, label='  S1')
    pypolar.visualization.draw_stokes_poincare(S2, ax, label='  S2')
    pypolar.visualization.join_stokes_poincare(S1, S2, ax, lw=2, ls=':', color='orange')
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.animation as animation
from mpl_toolkits.mplot3d import Axes3D

import pypolar.fresnel
import pypolar.mueller
import pypolar.jones

plt.rcParams["animation.html"] = "jshtml"

__all__ = ('draw_jones_field',
           'draw_jones_animated',
           'draw_jones_ellipse',
           'draw_stokes_ellipse',
           'draw_stokes_field',
           'draw_stokes_animated',
           'draw_empty_sphere',
           'draw_jones_poincare',
           'draw_stokes_poincare',
           'join_jones_poincare',
           'join_stokes_poincare'
           )

def _draw_optical_axis_3d(J, ax, last=4 * np.pi):
    """
    Draw the optical axis in a 3D plot.

    Args:
        J:    Jones vector
        ax:   matplotlib axis to use
        last: length of optical axis
    """
    h_amp, v_amp = abs(J)
    the_max = max(h_amp, v_amp) * 1.1

    ax.plot([0, last], [0, 0], [0, 0], 'k')
    ax.plot([0, 0], [-the_max, the_max], [0, 0], 'g')
    ax.plot([0, 0], [0, 0], [-the_max, the_max], 'b')
    ax.text(0, 0, 1, "y", ha="center")
    ax.text(0, 1, 0, "x", va="center")
    ax.text(last*1.05, 0, 0, "z", va="center")


def _draw_h_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the horizontal electric field in a 3D plot.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = np.abs(J[0]) * np.cos(t + offset - np.angle(J[0]))
    z = 0
    ax.plot(x, y, z, ':g')


def _draw_v_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the vertical electric field in a 3D plot.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = 0 * t
    z = np.abs(J[1]) * np.cos(t + offset - np.angle(J[1]))
    ax.plot(x, y, z, ':b')


def _draw_total_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the total electric field in a 3D plot.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = np.abs(J[0]) * np.cos(t + offset - np.angle(J[0]))
    z = np.abs(J[1]) * np.cos(t + offset - np.angle(J[1]))
    ax.plot(x, y, z, 'r')


def _draw_projected_vector_3d(J, ax, offset):
    """
    Draw the projection vector of the polarization field in 3D.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """
    y = np.abs(J[0]) * np.cos(offset - np.angle(J[0]))
    z = np.abs(J[1]) * np.cos(offset - np.angle(J[1]))

    x1, y1, z1 = 0, y, 0
    x2, y2, z2 = 0, y, z
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'g--')

    x1, y1, z1 = 0, 0, z
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'b--')

    x1, y1, z1 = 0, 0, 0
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'r')
    ax.scatter([0], [y], [z], marker='o', color='red')


def _draw_3D_field(J, ax, offset):
    """
    Draw a representation of the polarization fields in 3D.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """
    _draw_optical_axis_3d(J, ax)
    _draw_h_field_3d(J, ax, offset)
    _draw_v_field_3d(J, ax, offset)
    _draw_total_field_3d(J, ax, offset)
    _draw_projected_vector_3d(J, ax, offset)

    ax.grid(False)
    ax.axis('off')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])


def _draw_2D_field(J, ax, offset):
    """
    Draw a simple 2D representation of the projected field.

    Also called a sectional pattern.

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """
    h_amp, v_amp = np.abs(J)
    h_phi, v_phi = np.angle(J)
    the_max = max(h_amp, v_amp) * 1.1

    ax.plot([-the_max, the_max], [0, 0], 'g')
    ax.plot([0, 0], [-the_max, the_max], 'b')

    t = np.linspace(0, 2 * np.pi, 100)
    x = h_amp * np.cos(t + offset - h_phi)
    y = v_amp * np.cos(t + offset - v_phi)
    ax.plot(x, y, 'k')

    x = h_amp * np.cos(offset - h_phi)
    y = v_amp * np.cos(offset - v_phi)
    ax.plot(x, y, 'ro')
    ax.plot([x, x], [0, y], 'g--')
    ax.plot([0, x], [y, y], 'b--')
    ax.plot([0, x], [0, y], 'r')

    ax.set_xlim(-the_max, the_max)
    ax.set_ylim(-the_max, the_max)
    ax.set_aspect('equal')
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.text(0, 1, "y", ha="center")
    ax.text(1, 0, "x", va="center")


def _animation_update(offset, J, ax1, ax2):
    """
    Draw the next animation frame.

    Args:
        offset: starting phase for drawings
        J:      Jones vector
        ax1:    matplotlib axis for 3D plot
        ax2:    matplotlib axis for 2D plot
    """
    ax1.clear()
    ax2.clear()
    _draw_3D_field(J, ax1, offset)
    _draw_2D_field(J, ax2, offset)
    return ax1, ax2


def draw_ellipse_axes(J, ax):
    """
    Draw the sectional pattern with ellipse labels.

    Args:
        J:  Jones vector
        ax: plot axis
    """
    Ex0, Ey0 = np.abs(J)
    phix, phiy = np.angle(J)

    alpha = pypolar.jones.ellipse_azimuth(J)
    a, b = pypolar.jones.ellipse_axes(J)

    t = np.linspace(0, 2 * np.pi, 100)
    xx = Ex0 * np.cos(t + phix)
    yy = Ey0 * np.cos(t + phiy)

    the_max = max(Ex0, Ey0) * 1.2

    ax.set_aspect('equal')
    ax.plot(xx, yy, 'b')

    # semi-major diameter
    dx = a * np.cos(alpha)
    dy = a * np.sin(alpha)
    ax.plot([0, dx], [0, dy], 'r')
    ax.text(dx/2, dy/2, '  a', color='red')
    ax.text(dx/5, dy/10, r'$\alpha$', va='center', ha='center')
    s = r'a=%.2f, b=%.2f, $\alpha$=%.2f°' % (a, b, np.degrees(alpha))
    ax.text(0, -1.15*the_max, s, ha='center')

    # semi-minor diameter
    alpha += np.pi/2
    dx = b * np.cos(alpha)
    dy = b * np.sin(alpha)
    ax.plot([0, dx], [0, dy], 'g')
    ax.text(dx/2, dy/2, '  b', color='green')
    s = r'b/a=%.2f, ' % (b/a)
    s += r'$\tan^{-1}(b/a)$=%.2f°' % np.degrees(pypolar.jones.ellipticity_angle(J))
    ax.text(0, -1.30*the_max, s, ha='center')

    # draw x and y axes
    ax.plot([0, 0], [-the_max, the_max], 'k')
    ax.plot([-the_max, the_max], [0, 0], 'k')
    ax.set_xlim(-the_max, the_max)
    ax.set_ylim(-the_max, the_max)
    ax.set_xticks([])
    ax.set_yticks([])

def draw_ellipse_Ex_Ey(J, ax):
    """
    Draw the sectional pattern with field labels.

    Args:
        J:  Jones vector
        ax: plot axis
    """
    Ex0, Ey0 = np.abs(J)
    phix, phiy = np.angle(J)

    t = np.linspace(0, 2 * np.pi, 100)
    xx = Ex0 * np.cos(t + phix)
    yy = Ey0 * np.cos(t + phiy)

    the_max = max(Ex0, Ey0) * 1.2
    ax.set_aspect('equal')
    ax.plot(xx, yy, 'b')
    ax.plot([-Ex0, -Ex0, Ex0, Ex0, -Ex0], [-Ey0, Ey0, Ey0, -Ey0, -Ey0], ':g')
    ax.plot([-Ex0, Ex0], [-Ey0, Ey0], ':r')
    ax.plot([0, 0], [-the_max, the_max], 'k')
    ax.plot([-the_max, the_max], [0, 0], 'k')
    ax.text(Ex0, 0, r' $E_{x0}$', va='bottom', ha='left')
    ax.text(-Ex0, 0, r'$-E_{x0} $', va='bottom', ha='right')
    ax.text(0, Ey0, r'$E_{y0}$', va='bottom', ha='left')
    ax.text(0, -Ey0, r'$-E_{y0}$', va='top', ha='left')
    ax.text(0, Ey0/5, r' $\psi$', va='bottom', ha='left')
    ax.set_xlim(-the_max, the_max)
    ax.set_ylim(-the_max, the_max)
    ax.set_xticks([])
    ax.set_yticks([])
    psi = np.degrees(np.arctan2(Ex0, Ey0))
    s = r'$E_{0x}$=%.2f, $E_{0y}$=%.2f, $\psi$=%.2f°' % (Ex0, Ey0, psi)
    ax.text(0, -1.15*the_max, s, ha='center')
    s = r'$\phi_x$=%.2f°, ' % np.degrees(phix)
    s += r'$\phi_y$=%.2f°, ' % np.degrees(phiy)
    s += r'$\phi_y-\phi_x$=%.2f°' % np.degrees(phiy-phix)
    ax.text(0, -1.30*the_max, s, ha='center')


def draw_jones_ellipse(J, simple=False):
    """
    Draw a 2D sectional pattern for a Jones vector.

    Args:
        J:      Jones vector
        simple: if True then just draw a simple ellipse plot
    """
    JJ = J
    if pypolar.jones.alternate_sign_convention:
        JJ = np.conjugate(J)

    if simple:
        Ex0, Ey0 = np.abs(JJ)
        phix, phiy = np.angle(JJ)
        the_max = max(Ex0, Ey0) * 1.2
        t = np.linspace(0, 2 * np.pi, 100)
        xx = Ex0 * np.cos(t + phix)
        yy = Ey0 * np.cos(t + phiy)
        ax = plt.gca()
        ax.set_xlim(-the_max, the_max)
        ax.set_ylim(-the_max, the_max)
        ax.set_aspect('equal')
        ax.axhline(0, color='black')
        ax.axvline(0, color='black')
        ax.plot(xx, yy, 'b')
        ax.plot([-Ex0, Ex0], [-Ey0, Ey0], ':r')
        ax.axis('off')
        ax.text(0, Ey0/5, r' $\psi$', va='bottom', ha='left')
        return

    plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[1, 1])
    ax1 = plt.subplot(gs[0])
    draw_ellipse_axes(JJ, ax1)
    ax2 = plt.subplot(gs[1])
    draw_ellipse_Ex_Ey(JJ, ax2)


def draw_stokes_ellipse(S):
    """
    Draw a 2D and 3D representation of the polarization.

    Args:
        S:      Stokes vector
    """
    J = pypolar.mueller.stokes_to_jones(S)
    draw_jones_ellipse(J)


def draw_jones_field(J, offset=0):
    """
    Draw 3D and 2D representations of the polarization field.

    Args:
        J:      Jones vector
        offset: starting point
    """
    plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

    ax1 = plt.subplot(gs[0], projection='3d')
    _draw_3D_field(J, ax1, offset)

    ax2 = plt.subplot(gs[1])
    _draw_2D_field(J, ax2, offset)


def draw_stokes_field(S, offset=0):
    """
    Draw a 2D and 3D representation of the polarization.

    Args:
        S:      Stokes vector
        offset: starting point
    """
    J = pypolar.mueller.stokes_to_jones(S)
    draw_jones_field(J, offset)


def draw_jones_animated(J, nframes=64):
    """
    Animate 3D and 2D representations of the polarization field.

    Args:
        J:      Jones vector
    """
    JJ = J
    if pypolar.jones.alternate_sign_convention:
        JJ = np.conjugate(J)

    fig = plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])
    ax1 = plt.subplot(gs[0], projection='3d')
    ax2 = plt.subplot(gs[1])

    ani = animation.FuncAnimation(fig, _animation_update,
                                  frames=np.linspace(0, -2 * np.pi, nframes),
                                  fargs=(JJ, ax1, ax2))
    plt.close()
    return ani


def draw_stokes_animated(S):
    """
    Draw animated 2D and 3D representations of the polarization.

    Args:
        S:      Stokes vector
    """
    J = pypolar.mueller.stokes_to_jones(S)
    ani = draw_jones_animated(J)
    return ani


def draw_empty_sphere(ax=None):
    """
    Plot an empty Poincare sphere.

    Args:
        ax: pyplot axis
    """
    if ax is None:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(111, projection='3d')

    ax.view_init(elev=30, azim=45)

    try:
        ax.set_box_aspect((1, 1, 1))
    except AttributeError:
        try:
            ax.set_aspect('equal')
        except NotImplementedError:
            pass

    u = np.radians(np.linspace(0, 360, 90))
    v = np.radians(np.linspace(0, 180, 90))
    zz = np.zeros_like(u)

    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones_like(u), np.cos(v))

    ax.plot_surface(x, y, z, alpha=0.1, color='blue')

    # draw circumferences
    plt.plot(np.sin(u), np.cos(u), 0, 'k', lw=0.5)
    plt.plot(np.sin(u), zz, np.cos(u), 'k', lw=0.5)
    plt.plot(zz, np.sin(u), np.cos(u), 'k', lw=0.5)

    # draw x,y,z axes
    plt.plot([-1, 1], [0, 0], [0, 0], 'k--', lw=1, alpha=0.5)
    plt.plot([0, 0], [-1, 1], [0, 0], 'k--', lw=1, alpha=0.5)
    plt.plot([0, 0], [0, 0], [-1, 1], 'k--', lw=1, alpha=0.5)

    # label directions
    ax.text(1.15, 0, 0, '0°', fontsize=12, color='black', ha='center')
    ax.text(0, 1.25, 0, '45°', fontsize=12, color='black', ha='center')
    ax.text(0, 0, 1.15, 'RCP', fontsize=12, color='black', ha='center')
    ax.text(0, 0, -1.15, 'LCP', fontsize=12, color='black', ha='center')
    ax.text(-1.15, 0, 0, '90°', fontsize=12, color='black', ha='center')

    # Stokes parameters
    ax.set_xlabel('S₁', fontsize=14, labelpad=-10)
    ax.set_ylabel('S₂', fontsize=14, labelpad=-10)
    ax.set_zlabel('S₃', fontsize=14, labelpad=-10)

    # Hide grid and ticks
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])

def great_circle_points(ax, ay, az, bx, by, bz):
    """
    Create a list of points along the great circle between a and b.

    The great circle is assumed to lie on the unit sphere with center at (0,0,0)

    The points a=(ax,ay,az) and b=(bx,by,bz) are the beginning and end of the arc.

    Algorithm is from https://www.physicsforums.com/threads/571535
    """
    delta = np.arccos(ax*bx + ay*by + az*bz)
    psi = np.linspace(0, delta)
    sinpsi = np.sin(psi)
    cospsi = np.cos(psi)
    sindelta = np.sin(delta)

    # handle case when delta=0° or 180°
    if sindelta == 0:
        sindelta = 1e-5
    elif abs(sindelta) < 1e-5:
        sindelta = 1e-5 * np.sign(sindelta)

    x = cospsi * ax + sinpsi * ((az**2 + ay**2)*bx - (az*bz+ay*by)*ax)/sindelta
    y = cospsi * ay + sinpsi * ((az**2 + ax**2)*by - (az*bz+ax*bx)*ay)/sindelta
    z = cospsi * az + sinpsi * ((ay**2 + ax**2)*bz - (ay*by+ax*bx)*az)/sindelta
    return x, y, z

def spherical_angles(x, y, z):
    """Azimuth and elevation for a point on a sphere."""
    phi = np.arctan2(y, x)
    theta = np.arctan2(np.sqrt(x*x+y*y), z)
    return phi, theta

def draw_stokes_poincare(S, ax=None, label=None, **kwargs):
    """Plot single point on Poincaré sphere."""
    if ax is None:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(111, projection='3d')
        draw_empty_sphere(ax)

    SS = np.sqrt(S[1]**2+S[2]**2+S[3]**2)
    x = S[1]/SS
    y = S[2]/SS
    z = S[3]/SS

    plot_keys = ['lineweight', 'color', 'linestyle', 'markersize']
    plot_args = dict((k, kwargs[k]) for k in plot_keys if k in kwargs)
    ax.plot([x], [y], [z], 'o', **plot_args)

    if not label is None:
        text_keys = ['fontsize', 'ha', 'color', 'va']
        text_args = dict((k, kwargs[k]) for k in text_keys if k in kwargs)
        ax.text(x, y, z, label, **text_args)

def draw_jones_poincare(J, ax=None, label=None, **kwargs):
    """Plot single point on Poincaré sphere."""
    S = pypolar.jones.jones_to_stokes(J)
    draw_stokes_poincare(S, ax=ax, label=label, **kwargs)

def join_stokes_poincare(S1, S2, ax=None, **kwargs):
    """Plot arc joining two Stokes vectors on Poincaré sphere."""
    if ax is None:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(111, projection='3d')
        draw_empty_sphere(ax)

    SS1 = np.sqrt(S1[1]**2+S1[2]**2+S1[3]**2)
    SS2 = np.sqrt(S2[1]**2+S2[2]**2+S2[3]**2)
    x, y, z = great_circle_points(S1[1]/SS1, S1[2]/SS1, S1[3]/SS1, S2[1]/SS2, S2[2]/SS2, S2[3]/SS2)
    ax.plot(x, y, z, **kwargs)

def join_jones_poincare(J1, J2, ax=None, **kwargs):
    """Plot arc joining two Jones vectors on Poincaré sphere."""
    S1 = pypolar.jones.jones_to_stokes(J1)
    S2 = pypolar.jones.jones_to_stokes(J2)
    join_stokes_poincare(S1, S2, ax=ax, **kwargs)



# pylint: disable=invalid-name

"""
Useful functions for calculating light interaction at planar boundaries.

The underlying assumptions are that there a two semi-infinite media with a
planar interface.  For convenience, assume the light is incident from the top.
The upper medium is characterized by a purely real index of refraction `n_i` which
has a default value of 1.  The lower medium is characterized by a complex index
of refraction `m = n - n * kappa * 1j`.  Note that `pypolar` assumes the sign
of the imaginary part of the index of refraction is negative.

The Fresnel equations assume that the electric field has been decomposed into
fields relative to the plane of incidence (a plane defined by the incoming
light direction and the normal to the surface).

The incidence angle is measured from the normal to the surface and is measured
in radians.

To Do::
    * Make sure routines work for arrays of m or of theta_i
    * fail for positive imaginary refractive indices
    * fail for out-of-range angles to catch degrees/radians error

Scott Prahl
Apr 2020
"""

import numpy as np

__all__ = ('brewster',
           'critical',
           'r_par_amplitude',
           'r_per_amplitude',
           't_par_amplitude',
           't_per_amplitude',
           'R_par',
           'R_per',
           'T_par',
           'T_per',
           'R_unpolarized',
           'T_unpolarized'
           )

def brewster(m, n_i=1):
    """
    Brewster's angle for an interface.

    Args:
        m: index of refraction of the outgoing medium [-]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        Brewster's angle from normal to surface    [radians]
    """
    return np.arctan2(m, n_i)


def critical(m, n_i=1):
    """
    Critical angle for total internal reflection at interface.

    Args:
        m: index of refraction of the outgoing medium [-]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        critical angle from normal to surface    [radians]
    """
    return np.arcsin(m/n_i)


def r_par_amplitude(m, theta_i, n_i=1):
    """
    Reflected fraction of parallel-polarized field at an interface.

    This is the fraction of the incident electric field reflected at the
    interface between two semi-infinite media. The incident field is assumed
    to be polarized parallel (p) to the plane of incidence (transverse magnetic
    or TM field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        reflected fraction of parallel field                  [-]
    """
    m2 = (m/n_i)**2
    c = m2 * np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex) # = m*cos(theta_t)
    if m.imag == 0:  # choose right branch for dielectrics
        d = np.conjugate(d)
    rp = (c - d) / (c + d)
    return np.real_if_close(rp)


def r_per_amplitude(m, theta_i, n_i=1):
    """
    Reflected fraction of perpendicular-polarized field at an interface.

    This is the fraction of the incident electric field reflected at the
    interface between two semi-infinite media. The incident field is assumed
    to be polarized perpendicular (s, or senkrecht) to the plane of incidence
    (transverse electric or TE field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        reflected fraction of perpendicular field [-]
    """
    m2 = (m/n_i)**2
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex) # = m*cos(theta_t)
    if m.imag == 0:  # choose right branch for dielectrics
        d = np.conjugate(d)
    rs = (c - d) / (c + d)
    return np.real_if_close(rs)


def t_par_amplitude(m, theta_i, n_i=1):
    """
    Find the transmitted fraction of parallel-polarized field through an interface.

    This is the fraction of the incident electric field transmitted through the
    interface between two semi-infinite media. The incident field is assumed
    to be polarized parallel (p) to the plane of incidence (transverse magnetic
    or TM field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        transmitted fraction of parallel field                [-]
    """
    m2 = (m/n_i)**2
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex) # = m*cos(theta_t)
    if m.imag == 0:  # choose right branch for dielectrics
        d = np.conjugate(d)
    tp = 2 * c * (m/n_i) / (m2 * c + d)
    return np.real_if_close(tp)


def t_per_amplitude(m, theta_i, n_i=1):
    """
    Return the transmitted fraction of perpendicular-polarized field through an interface.

    This is the fraction of the incident electric field transmitted through the
    interface between two semi-infinite media. The incident field is assumed
    to be polarized perpendicular (s, or senkrecht) to the plane of incidence
    (transverse electric or TE field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        transmitted fraction of perpendicular field [-]
    """
    m2 = (m/n_i)**2
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex) # = m*cos(theta_t)
    if m.imag == 0:  # choose right branch for dielectrics
        d = np.conjugate(d)
    ts = 2 * d / (m/n_i)/ (c + d)
    return np.real_if_close(ts)


def R_par(m, theta_i, n_i=1):
    """
    Reflected fraction of parallel-polarized optical power by an interface.

    The reflected fraction of incident power (or flux) assuming that
    the electric field of the incident light is polarized parallel (p) to the
    plane of incidence (transverse magnetic or TM electric field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        reflected fraction of parallel-polarized irradiance [-]
    """
    return abs(r_par_amplitude(m, theta_i, n_i))**2


def R_per(m, theta_i, n_i=1):
    """
    Return the fraction of perpendicular-polarized optical power reflectedby an interface.

    The fraction of the incident power (or flux) reflected at the
    interface between two semi-infinite media. The incident light is assumed
    to be polarized perpendicular (s, or senkrecht) to the plane of incidence
    (transverse electric or TE field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        reflected fraction of perpendicular-polarized irradiance [-]
    """
    return abs(r_per_amplitude(m, theta_i, n_i))**2


def T_par(m, theta_i, n_i=1):
    """
    Return the transmitted fraction of parallel-polarized optical power through an interface.

    The transmitted fraction of incident power (or flux) assuming that
    the electric field of the incident light is polarized parallel (p) to the
    plane of incidence (transverse magnetic or TM electric field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        transmitted fraction of parallel-polarized irradiance [-]
    """
    m2 = (m/n_i)**2
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex)
    tp = 2 * c * (m/n_i) / (m2 * c + d)
    return np.abs(d / c * abs(tp)**2)


def T_per(m, theta_i, n_i=1):
    """
    Return the transmitted fraction of perpendicular-polarized optical power through an interface.

    The transmitted fraction of the incident power (or flux) through the
    interface between two semi-infinite media. The incident light is assumed
    to be polarized perpendicular (s, or senkrecht) to the plane of incidence
    (transverse electric or TE field).

    The index of refraction for medium of the incoming field defaults to 1, but
    can be set any real value. The medium of the outgoing field is characterized
    by an index of refraction that may be complex.

    Args:
        m: complex index of refraction of the outgoing medium [-]
        theta_i: angle incident from normal to surface        [radians]
        n: real index of refraction of the incoming medium    [-]
    Returns:
        transmitted fraction of perpendicular-polarized irradiance [-]
    """
    m2 = (m/n_i)**2
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m2 - s * s, dtype=np.complex)
    ts = 2 * c / (c + d)
    return np.abs(d / c * abs(ts)**2)


def R_unpolarized(m, theta_i, n_i=1):
    """
    Fraction of unpolarized light that is reflected.

    Calculate reflection fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (R_par(m, theta_i, n_i) + R_per(m, theta_i, n_i)) / 2


def T_unpolarized(m, theta_i, n_i=1):
    """
    Fraction of unpolarized light that is transmitted.

    Calculate transmitted fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (T_par(m, theta_i, n_i) + T_per(m, theta_i, n_i)) / 2



# pylint: disable=invalid-name
# pylint: disable=no-member

"""
Symbolic manipulation of polarization using the Stokes/Mueller calculus.

The routines are broken up into four groups: (1) creating Stokes vectors, (2)
creating Mueller matrix operators, (3) interpretation, and (4) conversion.

Functions to create Stokes vectors::

    stokes_linear(angle)
    stokes_left_circular()
    stokes_right_circular()
    stokes_horizontal()
    stokes_vertical()
    stokes_unpolarized()
    stokes_elliptical(DOP, azimuth, ellipticity)
    stokes_ellipsometry(tanpsi, Delta)

Functions to create Mueller matrix operators::

    op_linear_polarizer(angle)
    op_retarder(fast_axis_angle, phase_delay)
    op_attenuator(optical_density)
    op_mirror()
    op_rotation(angle)
    op_quarter_wave_plate(fast_axis_angle)
    op_half_wave_plate(fast_axis_angle)
    op_fresnel_reflection(index_of_refraction, incidence_angle)
    op_fresnel_transmission(index_of_refraction, incidence_angle)

Functions to interpret Stokes vectors::

    intensity(stokes_vector)
    degree_of_polarization(stokes_vector)
    ellipse_orientation(stokes_vector)
    ellipse_ellipticity(stokes_vector)
    ellipse_axes(stokes_vector)

Functions to convert::

    stokes_to_jones(stokes_vector)
    mueller_to_jones(mueller_matrix)
"""

import sympy
import pypolar.jones

import pypolar.sym_fresnel as sym_fresnel

__all__ = ('op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'stokes_linear',
           'stokes_left_circular',
           'stokes_right_circular',
           'stokes_horizontal',
           'stokes_vertical',
           'stokes_unpolarized',
           'intensity',
           'degree_of_polarization',
           'ellipse_orientation',
           'ellipse_ellipticity',
           'ellipse_axes',
           'stokes_to_jones',
           'mueller_to_jones')


def op_linear_polarizer(theta):
    """
    Mueller matrix operator for a rotated linear polarizer.

    The polarizer is rotated about a normal to its surface.

    Args:
        theta: rotation angle measured from the horizontal plane [radians]
    """
    C2 = sympy.cos(2 * theta)
    S2 = sympy.sin(2 * theta)
    lp = sympy.Matrix([[1, C2, S2, 0],
                       [C2, C2**2, C2 * S2, 0],
                       [S2, C2 * S2, S2 * S2, 0],
                       [0, 0, 0, 0]])
    return 0.5 * lp


def op_retarder(theta, delta):
    """
    Mueller matrix operator for a rotated optical retarder.

    The retarder is rotated about a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
        delta: phase delay introduced between fast and slow-axes         [radians]
    """
    C2 = sympy.cos(2 * theta)
    S2 = sympy.sin(2 * theta)
    C = sympy.cos(delta)
    S = sympy.sin(delta)
    ret = sympy.Matrix([[1, 0, 0, 0],
                        [0, C2**2 + C * S2**2, (1 - C) * S2 * C2, -S * S2],
                        [0, (1 - C) * C2 * S2, S2**2 + C * C2**2, S * C2],
                        [0, S * S2, -S * C2, C]])
    return ret


def op_attenuator(t):
    """
    Mueller matrix operator for an optical attenuator.

    Args:
        t: fraction of light getting through attenuator  [---]
    """
    att = sympy.Matrix([[t, 0, 0, 0],
                        [0, t, 0, 0],
                        [0, 0, t, 0],
                        [0, 0, 0, t]])
    return att


def op_mirror():
    """Mueller matrix operator for a perfect mirror."""
    mir = sympy.Matrix([[1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, -1, 0],
                        [0, 0, 0, -1]])
    return mir


def op_rotation(theta):
    """
    Mueller matrix operator to rotate light about the optical axis.

    Args:
        theta: rotation angle  [radians]
    """
    C2 = sympy.cos(2 * theta)
    S2 = sympy.sin(2 * theta)
    rot = sympy.Matrix([[1, 0, 0, 0],
                        [0, C2, S2, 0],
                        [0, -S2, C2, 0],
                        [0, 0, 0, 1]])
    return rot


def op_quarter_wave_plate(theta):
    """
    Mueller matrix operator for a rotated quarter-wave plate.

    The QWP is rotated about a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
    """
    C2 = sympy.cos(2 * theta)
    S2 = sympy.sin(2 * theta)
    qwp = sympy.Matrix([[1, 0, 0, 0],
                        [0, C2**2, C2 * S2, -S2],
                        [0, C2 * S2, S2 * S2, C2],
                        [0, S2, -C2, 0]])
    return qwp


def op_half_wave_plate(theta):
    """
    Mueller matrix operator for rotated half-wave plate.

    The HWP is rotated about a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
    """
    C2 = sympy.cos(2 * theta)
    S2 = sympy.sin(2 * theta)
    qwp = sympy.Matrix([[1, 0, 0, 0],
                        [0, C2**2 - S2**2, 2 * C2 * S2, 0],
                        [0, 2 * C2 * S2, S2 * S2 - C2**2, 0],
                        [0, 0, 0, -1]])
    return qwp


def op_fresnel_reflection(m, theta):
    """
    Mueller matrix operator for Fresnel reflection.

    Convert from the Jones operator to ensure that phase
    change are handled properly.

    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        4x4 Fresnel reflection operator       [-]
    """
    J = pypolar.jones.op_fresnel_reflection(m, theta)
    R = pypolar.jones.jones_op_to_mueller_op(J)
    return R


def op_fresnel_transmission(m, theta):
    """
    Mueller matrix operator for Fresnel transmission.

    Unclear if phase changes are handled properly.  See Collett,
    "Mueller-Stokes Matrix Formulation of Fresnel's Equations,"
    Am. J. Phys. 39, 517 (1971).

    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        4x4 Fresnel transmission operator         [-]
    """
    tau_p = sym_fresnel.T_par(m, theta)
    tau_s = sym_fresnel.T_per(m, theta)
    a = tau_s + tau_p
    b = tau_s - tau_p
    c = 2 * sympy.sqrt(tau_s*tau_p)
    mat = sympy.Matrix([[a, b, 0, 0],
                        [b, a, 0, 0],
                        [0, 0, c, 0],
                        [0, 0, 0, c]])
    return 0.5 * mat


def stokes_linear(theta):
    """
    Stokes vector for linear polarized light at angle.

    Args:
        m :     complex index of refraction       [-]
        theta : angle from horizontal plane      [radians]
    """
    return sympy.Matrix([1, sympy.cos(2*theta), sympy.sin(2*theta), 0])


def stokes_right_circular():
    """Stokes vector for right circular polarized light."""
    return sympy.Matrix([1, 0, 0, 1])


def stokes_left_circular():
    """Stokes vector for left circular polarized light."""
    return sympy.Matrix([1, 0, 0, -1])


def stokes_horizontal():
    """Stokes vector for horizontal polarized light."""
    return sympy.Matrix([1, 1, 0, 0])


def stokes_vertical():
    """Stokes vector for vertical polarized light."""
    return sympy.Matrix([1, -1, 0, 0])


def stokes_unpolarized():
    """Stokes vector for unpolarized light."""
    return sympy.Matrix([1, 0, 0, 0])


def intensity(S):
    """Return the intensity."""
    return S[0]


def degree_of_polarization(S):
    """Return the degree of polarization."""
    return S[0]/sympy.sqrt(S[1]**2+S[2]**2+S[3]**2)


def ellipse_orientation(S):
    """
    Return orientation of the polarization ellipse.

    The orientation is the angle between the major semi-axis and the x-axis of
    the polarization ellipse (often represented by psi).
    """
    return 1/2 * sympy.arctan2(S[2], S[1])


def ellipse_ellipticity(S):
    """
    Return ellipticity of the polarization ellipse.

    The ellipticity of the polarization ellipse (often
    represented by chi)
    """
    return 1/2 * sympy.arcsin(S[3]/S[0])


def ellipse_axes(S):
    """
    Return the semi-major and semi-minor axes.

    These are the axes of the polarization ellipse.
    """
    absL = sympy.sqrt(S[1]**2 + S[2]**2)
    A = sympy.sqrt((S[0] + absL)/2)
    B = sympy.sqrt((S[0] - absL)/2)
    return A, B


def stokes_to_jones(S):
    """
    Convert a Stokes vector to a Jones vector.

    This conversion loses some of the information in the Stokes
    vector because the unpolarized fraction is lost.  Furthermore,
    since the Jones vector can differ by an arbitrary phase,
    the phase is chosen to make the horizontal component real.

    Args:
        S : a Stokes vector

    Returns:
         a corresponding Jones vector
    """
    # Calculate the degree of polarization
    p = sympy.sqrt(S[1]**2 + S[2]**2 + S[3]**2) / S[0]

    # Normalize the Stokes parameters (first one will be 1, of course)
    Q = S[1] / (S[0] * p)
    U = S[2] / (S[0] * p)
    V = S[3] / (S[0] * p)

    # the Jones components
    A = sympy.sqrt((1 + Q) / 2)
    if A == 0:
        B = 1
    else:
        B = sympy.complex(U, -V) / (2 * A)

    # put them together in a vector with the amplitude of the polarized part
    return sympy.sqrt(S[0] * p) * sympy.Matrix([A, B])


def mueller_to_jones(M):
    """
    Convert a Mueller matrix to a Jones matrix.

    Theocaris, Matrix Theory of Photoelasticity, eqns 4.70-4.76, 1979

    Args:
        M : a 4x4 Mueller matrix

    Returns:
         the corresponding 2x2 Jones matrix
    """
    A = sympy.empty((2, 2))
    A[0, 0] = sympy.sqrt((M[0, 0]+M[0, 1]+M[1, 0]+M[1, 1])/2)
    A[0, 1] = sympy.sqrt((M[0, 0]+M[0, 1]-M[1, 0]-M[1, 1])/2)
    A[1, 0] = sympy.sqrt((M[0, 0]-M[0, 1]+M[1, 0]-M[1, 1])/2)
    A[1, 1] = sympy.sqrt((M[0, 0]-M[0, 1]-M[1, 0]+M[1, 1])/2)

    theta = sympy.empty((2, 2))
    theta[0, 0] = 0
    theta[0, 1] = -sympy.arctan2(M[0, 3]+M[1, 3], M[0, 2]+M[1, 2])
    theta[1, 0] = sympy.arctan2(M[3, 0]+M[3, 1], M[2, 0]+M[2, 1])
    theta[1, 1] = sympy.arctan2(M[3, 2]-M[2, 3], M[2, 2]+M[3, 3])

    return A*sympy.exp(1j*theta)



# pylint: disable=invalid-name
# pylint: disable=bare-except
# pylint: disable=global-statement
# pep257: disable=D401

"""
Useful routines for managing polarization with the Jones calculus.

The routines are broken into four broad categories: creating Jones vectors,
creating Jones Matrices, interpreting Jones Vectors, and converting to the
Mueller/Stokes matrix calculus.

Creating Jones vectors for specific polarization states::

    * field_linear(angle)
    * field_left_circular()
    * field_right_circular()
    * field_horizontal()
    * field_vertical()
    * field_ellipsometry(tanpsi, Delta)
    * field_elliptical(azimuth, elliptic_angle)

Creating Jones Matrices for polarizing elements::

    * op_linear_polarizer(angle)
    * op_retarder(fast_axis_angle, retardance)
    * op_attenuator(optical_density)
    * op_mirror()
    * op_rotation(angle)
    * op_quarter_wave_plate(fast_axis_angle)
    * op_half_wave_plate(fast_axis_angle)
    * op_fresnel_reflection(index_of_refraction, angle)
    * op_fresnel_transmission(index_of_refraction, angle)

Interpreting the polarization state::

    * use_alternate_convention(boolean)
    * interpret(jones_vector)
    * intensity(jones_vector)
    * phase(jones_vector)
    * ellipse_azimuth(jones_vector)
    * ellipse_axes(jones_vector)
    * ellipticity(jones_vector)
    * ellipticity_angle(jones_vector)
    * amplitude_ratio(jones_vector)
    * amplitude_ratio_angle(jones_vector)
    * polarization_variable(jones_vector)

Converting to Mueller formalism::

    * jones_op_to_mueller_op(jones_matrix)
    * jones_to_stokes(jones_vector)

To Do::
    * modify interpret() when phase difference differs by more than 2pi
    * improve interpret() to give angle for elliptical polarization
    * test everything with non-unity amplitudes
    * finish normalize
"""

import numpy as np
import pypolar.fresnel

__all__ = ('use_alternate_convention',
           'op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'field_linear',
           'field_left_circular',
           'field_right_circular',
           'field_horizontal',
           'field_vertical',
           'field_ellipsometry',
           'field_elliptical',
           'interpret',
           'intensity',
           'phase',
           'ellipse_azimuth',
           'ellipse_axes',
           'ellipticity',
           'ellipticity_angle',
           'amplitude_ratio',
           'amplitude_ratio_angle',
           'polarization_variable',
           'jones_op_to_mueller_op',
           'jones_to_stokes'
           )

alternate_sign_convention = False

def use_alternate_convention(state):
    """
    Change sign convention used for Jones calculus.

    Read the documentation about the different conventions possible.

    The default convention is to assume the wave function is
    represented by exp(j*(omega*t-k*z)) and that the perspective
    when viewing a sectional pattern is to look back along the
    optical axis towards the source.  This is the most commonly used
    convention, but there are noteable exceptions (Wikipedia, Fowler, and
    Hecht (sometimes).

    Call this function once at the beginning and everything should
    be just fine.
    """
    global alternate_sign_convention
    alternate_sign_convention = state

def op_linear_polarizer(theta):
    """
    Jones matrix operator for a rotated linear polarizer.

    The polarizer has been rotated around a normal to its surface.

    Args:
        theta: rotation angle measured from the horizontal plane [radians]
    """
    return np.array([[np.cos(theta)**2, np.sin(theta) * np.cos(theta)],
                     [np.sin(theta) * np.cos(theta), np.sin(theta)**2]])


def op_retarder(theta, delta):
    """
    Jones matrix operator for an rotated optical retarder.

    The retarder has been rotated around a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
        delta: phase delay introduced between fast and slow-axes         [radians]
    """
    if alternate_sign_convention:
        theta *= -1
    P = np.exp(+delta / 2 * 1j)
    Q = np.exp(-delta / 2 * 1j)
    D = np.sin(delta / 2) * 2j
    C = np.cos(theta)
    S = np.sin(theta)
    retarder = np.array([[C * C * P + S * S * Q, C * S * D],
                         [C * S * D, C * C * Q + S * S * P]])
    if alternate_sign_convention:
        return np.conjugate(retarder)
    return retarder


def op_attenuator(t):
    """
    Jones matrix operator for an isotropic optical attenuator.

    The transmittance t=I/I_0 is the fraction of light getting
    through the attenuator or absorber.

    Args:
        t: fraction of intensity getting through attenuator  [---]
    """
    f = np.sqrt(t)
    return np.array([[f, 0], [0, f]])


def op_mirror():
    """Jones matrix operator for a perfect mirror."""
    return np.array([[1, 0], [0, -1]])


def op_rotation(theta):
    """
    Jones matrix operator to rotate light around the optical axis.

    Args:
        theta : angle of rotation around optical axis  [radians]
    Returns:
        2x2 matrix of the rotation operator           [-]
    """
    return np.array([[np.cos(theta), np.sin(theta)],
                     [-np.sin(theta), np.cos(theta)]])


def op_quarter_wave_plate(theta):
    """
    Jones matrix operator for an rotated quarter-wave plate.

    The QWP had been rotated around a normal to its surface.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the quarter-wave plate operator     [-]
    """
    return op_retarder(theta, np.pi / 2)


def op_half_wave_plate(theta):
    """
    Jones matrix operator for a rotated half-wave plate.

    The half wave plate has been rotated around a normal to
    the surface of the plate.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the half-wave plate operator     [-]
    """
    return op_retarder(theta, np.pi)


def op_fresnel_reflection(m, theta):
    """
    Jones matrix operator for Fresnel reflection at angle theta.

    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        2x2 matrix of the Fresnel transmission operator     [-]
    """
    return np.array([[pypolar.fresnel.r_par_amplitude(m, theta), 0],
                     [0, pypolar.fresnel.r_per_amplitude(m, theta)]])


def op_fresnel_transmission(m, theta):
    """
    Jones matrix operator for Fresnel transmission at angle theta.

    *** THIS IS ALMOST CERTAINLY WRONG ***

    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        2x2 Fresnel transmission operator           [-]
    """
    c = np.cos(theta)
    d = np.sqrt(m * m - np.sin(theta)**2, dtype=np.complex)
    if m.imag == 0:
        d = np.conjugate(d)
    a = np.sqrt(d/c)
    tper = pypolar.fresnel.t_per_amplitude(m, theta)
    tpar = pypolar.fresnel.t_par_amplitude(m, theta)
    return a*np.array([[tpar, 0], [0, tper]])


def field_linear(theta):
    """Jones vector for linear polarized light at angle theta from horizontal plane."""
    return np.array([np.cos(theta), np.sin(theta)]).T


def field_right_circular():
    """Jones Vector for right circular polarized light."""
    J = 1 / np.sqrt(2) * np.array([1, 1j])
    if alternate_sign_convention:
        return np.conjugate(J)
    return J


def field_left_circular():
    """Jones Vector for left circular polarized light."""
    J = 1 / np.sqrt(2) * np.array([1, -1j])
    if alternate_sign_convention:
        return np.conjugate(J)
    return J


def field_horizontal():
    """Jones Vector for horizontal polarized light."""
    return np.array([1, 0])


def field_vertical():
    """Jones Vector for vertical polarized light."""
    return np.array([0, 1])


def field_ellipsometry(tanpsi, Delta):
    """
    Jones vector for using ellipometer parameters.

    This creates a Jones vector for the specific set of ellipsometry
    parameters tanpsi and Delta.  See Fujiwara table 3.1 for example

    Args:
        tanpsi: abs(E_x/E_y)             [-]
        Delta: angle(E_x) - angle(E_y)   [radians]
    Returns:
        Jones vector with specified characteristics
    """
    psi = np.arctan(tanpsi)
    J = np.array([np.sin(psi)*np.exp(1j*Delta), np.cos(psi)]).T
    if alternate_sign_convention:
        return np.conjugate(J)
    return J


def field_elliptical(azimuth, elliptic_angle, phi_x=0, E_0=1):
    """
    Jones vector for elliptically polarized light.

    Uses Azzam's equation 1.75

    Args:
        azimuth: tilt angle of ellipse from x-axis        [radians]
        ellipticity_angle: arctan(minor-axis/major-axis)  [radians]
        phi_x: phase for E field in x-direction           [radians]
        E_0: amplitude of field
    Returns:
        Jones vector with specified characteristics
    """
    ce = np.cos(elliptic_angle)
    se = np.sin(elliptic_angle)
    ca = np.cos(azimuth)
    sa = np.sin(azimuth)

    J = E_0 * np.array([ca*ce-sa*se*1j, sa*ce+ca*se*1j])

    J *= np.exp(1j * (phi_x-np.angle(J[0])))

    if alternate_sign_convention:
        return np.conjugate(J)
    return J

def interpret(J):
    """
    Interpret a Jones vector.

    arg:
        J: A Jones vector (2x1) which may have complex entries

    Examples
    -------
    interpret([1, -1j]) --> "Right circular polarization"

    interpret([0.5, 0.5]) -->
                      "Linear polarization at 45.000000 degrees CCW from x-axis"

    interpret( np.array([exp(-1j*pi), exp(-1j*pi/3)]) ) -->
                "Left elliptical polarization, rotated with respect to the axes"
    """
    try:
        j1, j2 = J
    except:
        print("Jones vector must have two elements")
        return 0

    eps = 1e-12
    mag1, p1 = abs(j1), np.angle(j1)
    mag2, p2 = abs(j2), np.angle(j2)

    JJ = np.array([j1, j2])
    inten = intensity(JJ)
    phaze = np.degrees(phase(JJ))
    azi = np.degrees(ellipse_azimuth(JJ))
    ell = ellipticity(JJ)

    s = "Intensity is %.3f\n" % inten
    s += "Phase is %.1f°\n" % phaze

    if np.remainder(p1 - p2, np.pi) < eps:
        ang = np.arctan2(mag2, mag1) * 180 / np.pi
        return s + "Linear polarization at %f degrees CCW from x-axis" % ang

    if abs(mag1 - mag2) < eps:
        if abs(p1 - p2 - np.pi / 2) < eps:
            s += "Right circular polarization"
        elif p1 > p2:
            s += "Right elliptical polarization\n"
            s += "    ellipticity is %.1f°\n" % ell
            s += "    rotated %.1f° respect to the axes" % azi
        if (p1 - p2 + np.pi / 2) < eps:
            s += "Left circular polarization"
        elif p1 < p2:
            s += "Left elliptical polarization\n"
            s += "    ellipticity is %.1f°\n" % ell
            s += "    rotated %.1f° respect to the axes" % azi
    else:
        if p1 - p2 == np.pi / 2:
            s += "Right elliptical polarization, non-rotated"
        elif p1 > p2:
            s += "Right elliptical polarization\n"
            s += "    ellipticity is %.1f°\n" % ell
            s += "    rotated %.1f° respect to the axes" % azi
        if p1 - p2 == -np.pi / 2:
            s += "Left circular polarization, non-rotated"
        elif p1 < p2:
            s += "Left elliptical polarization\n"
            s += "    ellipticity is %.1f°\n" % ell
            s += "    rotated %.1f° respect to the axes" % azi
    return s


def normalize_vector(J):
    """
    Normalize a vector by dividing each part by common number.

    After normalization the magnitude should be equal to ~1.
    """
    norm = np.linalg.norm(J)
    if norm == 0:
        return J
    return J / norm


def normalize(J):
    """Normalize a vector."""
#    alpha = ellipse_azimuth(J)
#    gamma = phase(J)
    return J
#    return np.array([np.cos(R)*np.exp(-0.5j*gamma),np.cos(R)*np.exp(0.5j*gamma)])


def intensity(J):
    """Return the intensity."""
    inten = abs(J[..., 0])**2 + abs(J[..., 1])**2
    return inten


def phase(J):
    """Return the phase."""
    gamma = np.angle(J[..., 1]) - np.angle(J[..., 0])
    return gamma


def ellipse_azimuth(J):
    """
    Return the angle between the major semi-axis and the x-axis.

    The polarization ellipse is rotated by this angle (called
    the azimuth) relative to the laboratory frame.
    """
    Ex0, Ey0 = np.abs(J)
    delta = phase(J)
    numer = 2 * Ex0 * Ey0 * np.cos(delta)
    denom = Ex0**2 - Ey0**2
    psi = 0.5 * np.arctan2(numer, denom)
    return psi


def ellipse_azimuth2(J):
    """
    Return the angle between the major semi-axis and the x-axis.

    How does this differ from orientation above?
    """
    delta = phase(J)
    psi = ellipse_azimuth(J)
    chi = 0.5 * np.arcsin(np.sin(2 * psi) * np.sin(delta))
    return chi


def ellipse_axes(J):
    """
    Return the semi-major and semi-minor radii of the ellipse.

    Twice these values will be the semi-major or semi-minor diameters.
    """
    Ex0, Ey0 = np.abs(J)
    alpha = ellipse_azimuth(J)
    delta = phase(J)
    C = np.cos(alpha)
    S = np.sin(alpha)
    asqr = (Ex0 * C)**2 + (Ey0 * S)**2 + 2 * Ex0 * Ey0 * C * S * np.cos(delta)
    bsqr = (Ex0 * S)**2 + (Ey0 * C)**2 - 2 * Ex0 * Ey0 * C * S * np.cos(delta)
    a = np.sqrt(abs(asqr))
    b = np.sqrt(abs(bsqr))
    if a < b:
        return b, a
    return a, b


def ellipticity(J):
    """
    Return the ellipticity of the polarization ellipse.

    This is the ratio of semi-minor to semi-major radii.
    The ellipticity is a measure of the fatness of the ellipse.
    The ellipticity can be defined to always be positive.  However
    negative values can be used to indicate left-handed polarization.
    Thus the ellipticity will range from -1 to 0 to 1 as light moves from
    LCP to Linear Polarization to RCP.
    """
    a, b = ellipse_axes(J)
    if phase(J) < 0:
        return -b/a
    return b/a


def ellipticity_angle(J):
    """
    Return the ellipticity angle of the polarization ellipse.

    The tangent of this angle is the ratio of semi-minor:semi-major
    radii.  It is between -pi/4 ≤ angle ≤ pi/4.  Positive values
    are for right–handed ellipticity.  Negative values for left-handed
    ellipticity.
    """
    a, b = ellipse_axes(J)
    if abs(a) >= abs(b):
        epsilon = np.arctan2(b, a)
    else:
        epsilon = np.arctan2(a, b)

    if phase(J) < 0:
        return -epsilon
    return epsilon


def amplitude_ratio(J):
    """
    Return the ratio of electric fields.

    This is the amplitude of the vibrations along x measured
    relative to the amplitude along y.
    """
    Ex0, Ey0 = np.abs(J)
    if Ex0 == 0:
        return np.inf
    return Ey0/Ex0


def amplitude_ratio_angle(J):
    """
    Return the ratio of electric fields.

    The tangent of this angle is the ratio of electric fields in
    the y and x directions.
    """
    Ex0, Ey0 = np.abs(J)
    psi = np.arctan2(Ey0, Ex0)
    return psi


def polarization_variable(J):
    """
    Return the complex polarization variable, chi.

    This reduces the Jones vector to a single complex number and is useful
    when the amplitude and absolute-phase are of secondary interest.  These
    are eliminated and chi is representative of the polarization state
    in the complex plane.
    """
    return J[1]/J[0]


def poincare_point(J):
    """Return the point on the Poincaré sphere."""
    longitude = 2 * ellipse_azimuth(J)
    a, b = ellipse_axes(J)
    latitude = 2 * np.arctan2(b, a)
    return latitude, longitude


def jones_op_to_mueller_op(JJ):
    """
    Convert a complex 2x2 Jones matrix to a real 4x4 Mueller matrix.

    Hauge, Muller, and Smith, "Conventions and Formulas for Using the Mueller-
    Stokes Calculus in Ellipsometry," Surface Science, 96, 81-107 (1980)
    Args:
        J:      Jones matrix
    Returns:
        equivalent 4x4 Mueller matrix
    """
    if alternate_sign_convention:
        J = np.conjugate(JJ)
    else:
        J = JJ
    M = np.zeros(shape=[4, 4], dtype=np.complex)
    C = np.conjugate(J)
    M[0, 0] = J[0, 0] * C[0, 0] + J[0, 1] * C[0, 1] + \
              J[1, 0] * C[1, 0] + J[1, 1] * C[1, 1]
    M[0, 1] = J[0, 0] * C[0, 0] + J[1, 0] * C[1, 0] - \
              J[0, 1] * C[0, 1] - J[1, 1] * C[1, 1]
    M[0, 2] = J[0, 1] * C[0, 0] + J[1, 1] * C[1, 0] + \
              J[0, 0] * C[0, 1] + J[1, 0] * C[1, 1]
    M[0, 3] = 1j * (J[0, 1] * C[0, 0] + J[1, 1] * C[1, 0] -
                    J[0, 0] * C[0, 1] - J[1, 0] * C[1, 1])

    M[1, 0] = J[0, 0] * C[0, 0] + J[0, 1] * C[0, 1] - \
              J[1, 0] * C[1, 0] - J[1, 1] * C[1, 1]
    M[1, 1] = J[0, 0] * C[0, 0] - J[1, 0] * C[1, 0] - \
              J[0, 1] * C[0, 1] + J[1, 1] * C[1, 1]
    M[1, 2] = J[0, 0] * C[0, 1] + J[0, 1] * C[0, 0] - \
              J[1, 0] * C[1, 1] - J[1, 1] * C[1, 0]
    M[1, 3] = 1j * (J[0, 1] * C[0, 0] + J[1, 0] * C[1, 1] -
                    J[1, 1] * C[1, 0] - J[0, 0] * C[0, 1])

    M[2, 0] = J[0, 0] * C[1, 0] + J[1, 0] * C[0, 0] + \
              J[0, 1] * C[1, 1] + J[1, 1] * C[0, 1]
    M[2, 1] = J[0, 0] * C[1, 0] + J[1, 0] * C[0, 0] - \
              J[0, 1] * C[1, 1] - J[1, 1] * C[0, 1]
    M[2, 2] = J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] + \
              J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0]
    M[2, 3] = 1j * (-J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] -
                    J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0])

    M[3, 0] = 1j * (J[0, 0] * C[1, 0] + J[0, 1] * C[1, 1] -
                    J[1, 0] * C[0, 0] - J[1, 1] * C[0, 1])
    M[3, 1] = 1j * (J[0, 0] * C[1, 0] - J[0, 1] * C[1, 1] -
                    J[1, 0] * C[0, 0] + J[1, 1] * C[0, 1])
    M[3, 2] = 1j * (J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] -
                    J[1, 0] * C[0, 1] - J[1, 1] * C[0, 0])
    M[3, 3] = J[0, 0] * C[1, 1] - J[0, 1] * C[1, 0] - \
              J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0]
    MM = M.real / 2
    return MM

def _jones_to_stokes(J):
    """
    Convert Jones vector to Stokes vector.

    Args:
        J: Jones vector
    Returns:
        Stokes vector
    """
    Ex = abs(J[0])
    Ey = abs(J[1])
    phi = phase(J)

    S0 = Ex**2 + Ey**2
    S1 = Ex**2 - Ey**2
    S2 = 2*Ex*Ey*np.cos(phi)
    S3 = 2*Ex*Ey*np.sin(phi)
    return np.array([S0, S1, S2, S3])

def jones_to_stokes(J):
    """
    Convert Jones vector to Stokes vector.

    Args:
        J: Jones vector
    Returns:
        Stokes vector
    """
    if J.ndim == 1:
        return _jones_to_stokes(J)

    n, m = J.shape
    if m != 2:
        print("Wrong shape ... should be %dx2 not %dx%d" % (m, n, m))
        return None

    S = np.empty(shape=(n, 4), dtype=np.ndarray)
    for i, JJ in enumerate(J):
        S[i] = _jones_to_stokes(JJ)

    return S



from .fresnel import *
from .jones import *
from .mueller import *
from .visualization import *

from .sym_fresnel import *
from .sym_jones import *
from .sym_mueller import *

from .ellipsometry import *
from .gaertner import *



# pylint: disable=invalid-name
# pylint: disable=bare-except
"""
Useful basic routines for managing polarization with the Stokes/Mueller calculus.

The routines are broken up into four groups: (1) creating Stokes vectors, (2)
creating Mueller matrix operators, (3) interpretation, and (4) conversion.

Functions to create Stokes vectors::

    stokes_linear(angle)
    stokes_left_circular()
    stokes_right_circular()
    stokes_horizontal()
    stokes_vertical()
    stokes_unpolarized()
    stokes_elliptical(DOP, azimuth, ellipticity)
    stokes_ellipsometry(tanpsi, Delta)

Functions to create Mueller matrix operators::

    op_linear_polarizer(angle)
    op_retarder(fast_axis_angle, phase_delay)
    op_attenuator(optical_density)
    op_mirror()
    op_rotation(angle)
    op_quarter_wave_plate(fast_axis_angle)
    op_half_wave_plate(fast_axis_angle)
    op_fresnel_reflection(index_of_refraction, incidence_angle)
    op_fresnel_transmission(index_of_refraction, incidence_angle)

Functions to interpret Stokes vectors::

    intensity(stokes_vector)
    degree_of_polarization(stokes_vector)
    ellipse_orientation(stokes_vector)
    ellipse_ellipticity(stokes_vector)
    ellipse_axes(stokes_vector)
    interpret(stokes_vector)

Functions to convert::

    stokes_to_jones(stokes_vector)
    mueller_to_jones(mueller_matrix)
"""

import numpy as np
import pypolar.jones
import pypolar.fresnel

__all__ = ('op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'stokes_linear',
           'stokes_left_circular',
           'stokes_right_circular',
           'stokes_horizontal',
           'stokes_vertical',
           'stokes_unpolarized',
           'stokes_ellipsometry',
           'stokes_elliptical',
           'intensity',
           'degree_of_polarization',
           'ellipse_orientation',
           'ellipse_ellipticity',
           'ellipse_axes',
           'stokes_to_jones',
           'mueller_to_jones',
           'interpret')


def op_linear_polarizer(theta):
    """
    Mueller matrix operator for a rotated linear polarizer.

    The polarizer is rotated around a normal to its surface.

    Args:
        theta: rotation angle measured from the horizontal plane [radians]
    """
    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    lp = np.array([[1, C2, S2, 0],
                   [C2, C2**2, C2 * S2, 0],
                   [S2, C2 * S2, S2 * S2, 0],
                   [0, 0, 0, 0]])
    return 0.5 * lp


def op_retarder(theta, delta):
    """
    Mueller matrix operator for an rotated optical retarder.

    The retarder is rotated around a normal to its surface.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]
        delta: phase delay introduced between fast and slow-axes         [radians]
    """
    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    C = np.cos(delta)
    S = np.sin(delta)
    ret = np.array([[1, 0, 0, 0],
                    [0, C2**2 + C * S2**2, (1 - C) * S2 * C2, -S * S2],
                    [0, (1 - C) * C2 * S2, S2**2 + C * C2**2, S * C2],
                    [0, S * S2, -S * C2, C]])
    return ret


def op_attenuator(t):
    """
    Mueller matrix operator for an optical attenuator.

    Args:
        t : fraction of light getting through attenuator [---]
    """
    att = np.array([[t, 0, 0, 0],
                    [0, t, 0, 0],
                    [0, 0, t, 0],
                    [0, 0, 0, t]])
    return att


def op_mirror():
    """Mueller matrix operator for a perfect mirror."""
    mir = np.array([[1, 0, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, -1, 0],
                    [0, 0, 0, -1]])
    return mir


def op_rotation(theta):
    """
    Mueller matrix operator to rotate light around the optical axis.

    Args:
        theta: rotation angle  [radians]
    """
    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    rot = np.array([[1, 0, 0, 0],
                    [0, C2, S2, 0],
                    [0, -S2, C2, 0],
                    [0, 0, 0, 1]])
    return rot


def op_quarter_wave_plate(theta):
    """
    Mueller matrix operator for an quarter-wave plate.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]

    Returns:
        a Mueller matrix operator for the rotated quarter-wave plate.
    """
    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    qwp = np.array([[1, 0, 0, 0],
                    [0, C2**2, C2 * S2, -S2],
                    [0, C2 * S2, S2 * S2, C2],
                    [0, S2, -C2, 0]])
    return qwp


def op_half_wave_plate(theta):
    """
    Mueller matrix for a rotated half-wave plate.

    Args:
        theta: rotation angle between fast-axis and the horizontal plane [radians]

    Returns:
        a Mueller matrix operator for the rotated half-wave plate.
    """
    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    qwp = np.array([[1, 0, 0, 0],
                    [0, C2**2 - S2**2, 2 * C2 * S2, 0],
                    [0, 2 * C2 * S2, S2 * S2 - C2**2, 0],
                    [0, 0, 0, -1]])
    return qwp


def op_fresnel_reflection(m, theta):
    """
    Mueller matrix operator for Fresnel reflection at angle theta.

    These are based on Collett, Mueller-Stokes Matrix Formulation of Fresnel
    equation, Am. J Phys., 39, 1971.

    The changes in direction and detector orientation are included in the
    Mueller matrix.  See Clark, *Stellar Polarimetry*, Appendix A.

    Still needs sign testing for angles above Brewster's angle.

    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        4x4 Fresnel reflection Mueller matrix       [-]
    """
    rho_p = pypolar.fresnel.r_par_amplitude(m, theta)
    rho_s = pypolar.fresnel.r_per_amplitude(m, theta)
    a = abs(rho_s)**2 + abs(rho_p)**2
    b = abs(rho_s)**2 - abs(rho_p)**2
    c = 2 * rho_s * rho_p
    mat = np.array([[a, b, 0, 0],
                    [b, a, 0, 0],
                    [0, 0, c, 0],
                    [0, 0, 0, c]])
    return 0.5 * mat


def op_fresnel_transmission(m, theta):
    """
    Mueller matrix operator for Fresnel transmission at angle theta.

    These are based on Collett, Mueller-Stokes Matrix Formulation of Fresnel
    equation, Am. J Phys., 39, 1971.

    Still needs sign testing for angles above Brewster's angle.

    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        4x4 Fresnel transmission operator         [-]
    """
    tau_p = pypolar.fresnel.T_par(m, theta)
    tau_s = pypolar.fresnel.T_per(m, theta)
    a = tau_s + tau_p
    b = tau_s - tau_p
    c = 2 * np.sqrt(tau_s * tau_p)
    mat = np.array([[a, b, 0, 0],
                    [b, a, 0, 0],
                    [0, 0, c, 0],
                    [0, 0, 0, c]])
    return 0.5 * mat


def stokes_linear(theta):
    """Stokes vector for light polarized at angle theta from the horizontal plane."""
    if np.isscalar(theta):
        return np.array([1, np.cos(2*theta), np.sin(2*theta), 0])
    return np.array([np.ones_like(theta),
                     np.cos(2*theta),
                     np.sin(2*theta),
                     np.zeros_like(theta)]).T


def stokes_right_circular():
    """Stokes vector for right circular polarized light."""
    return np.array([1, 0, 0, 1])


def stokes_left_circular():
    """Stokes vector for left circular polarized light."""
    return np.array([1, 0, 0, -1])


def stokes_horizontal():
    """Stokes vector for horizontal polarized light."""
    return np.array([1, 1, 0, 0])


def stokes_vertical():
    """Stokes vector for vertical polarized light."""
    return np.array([1, -1, 0, 0])


def stokes_unpolarized():
    """Stokes vector for vertical polarized light."""
    return np.array([1, 0, 0, 0])


def stokes_ellipsometry(tanpsi, Delta):
    """
    Stokes vector using ellipsometer parameters.

    This creates a Stokes vector for the specific set of ellipsometry
    parameters tanpsi and Delta.  See Fujiwara table 3.1 for example.

    Args:
        tanpsi: abs(E_x/E_y)             [-]
        Delta: angle(E_x) - angle(E_y)   [radians]
    Returns:
        normalized Stokes vector with specified properties
    """
    psi = np.arctan(tanpsi)
    cp = np.cos(2*psi)
    sp = np.sin(2*psi)
    cd = np.cos(2*Delta)
    sd = np.sin(2*Delta)
    if np.isscalar(tanpsi) and np.isscalar(Delta):
        return np.array([1, -cp, sp*cd, -sp*sd])

    return np.array([np.ones_like(tanpsi), -cp, sp*cd, -sp*sd]).T


def stokes_elliptical(DOP, azimuth, ellipticity):
    """
    Stokes vector for partially polarized elliptically polarized light.

    Args:
        DOP: degree of polarization                     [-]
        azimuth: tilt of ellipse relative to horizontal [radians]
        ellipticity: ratio of minor to major axes       [-]
    Returns:
        normalized Stokes vector with specified properties
    """
    omega = np.arctan(ellipticity)
    cw = np.cos(2*omega)
    sw = np.sin(2*omega)
    ca = np.cos(2*azimuth)
    sa = np.sin(2*azimuth)
    if np.isscalar(DOP):
        unpolarized = np.array([1-DOP, 0, 0, 0])
        polarized = DOP * np.array([1, cw*ca, cw*sa, sw])
        return unpolarized + polarized

    unpolarized = np.array([np.ones_like(DOP)-DOP,
                            np.zeros_like(DOP),
                            np.zeros_like(DOP),
                            np.zeros_like(DOP)
                            ])
    polarized = DOP * np.array([np.ones_like(DOP), cw*ca, cw*sa, sw])
    return (unpolarized + polarized).T


def intensity(S):
    """Return the intensity."""
    if S.ndim == 1:
        return S[0]
    return S[..., 0]


def _degree_of_polarization(S):
    """Return the degree of polarization."""
    if S[0] == 0:
        return 0
    return np.sqrt(S[1]**2+S[2]**2+S[3]**2)/S[0]


def degree_of_polarization(S):
    """Return the degree of polarization."""
    if S.ndim == 1:
        return _degree_of_polarization(S)

    n, _ = S.shape
    dop = np.zeros(n)
    for i, SS in enumerate(S):
        dop[i] = _degree_of_polarization(SS)
    return dop


def ellipse_orientation(S):
    """
    Return the angle between the major semi-axis and the x-axis.

    The polarization ellipse is rotated by an angle from the
    laboratory frame.  This is that angle: often represented by psi.
    """
    return 1/2 * np.arctan2(S[..., 2], S[..., 1])


def ellipse_ellipticity(S):
    """
    Return the ellipticity of the polarization ellipse.

    This parameter is often represented by Chi.
    """
    return 1/2 * np.arcsin(S[..., 3]/S[..., 0])


def ellipse_axes(S):
    """Return the semi-major and semi-minor axes of the polarization ellipse."""
    absL = np.sqrt(S[..., 1]**2 + S[..., 2]**2)
    A = np.sqrt((S[..., 0] + absL)/2)
    B = np.sqrt((S[..., 0] - absL)/2)
    return A, B


def _stokes_to_jones(S):
    """
    Convert a Stokes vector to a Jones vector.

    The Jones vector can only represent the part of the Stokes vector that is
    polarized.  This fraction is calculated and represented as a Jones vector
    with its horizontal component represented as a real number.

    The sign convention for the Jones vector can be set by calling
    `pypolar.jones.use_alternate_convention(True)`.  The default is to assume that
    the field is represented by exp(j*omega*t-k*z).

    Inputs:
        S : a Stokes vector

    Returns:
         the Jones vector for
    """
    if S[0] == 0:
        return np.array([0, 0])

    # Fraction of intensity that is polarized
    Ip = np.sqrt(S[1]**2 + S[2]**2 + S[3]**2)

    # Normalize the remaining Stokes parameters to this fraction
    Q = S[1] / Ip
    U = S[2] / Ip
    V = S[3] / Ip

    # Amplitude of the polarized field
    E_0 = np.sqrt(Ip)

    # vertically polarized light has no E_x field
    if Q == -1:
        return np.array([0, E_0])

    # Assemble the Jones vector
    A = np.sqrt((1 + Q) / 2)
    J = E_0 * np.array([A, complex(U, V) / (2 * A)])

    if pypolar.jones.alternate_sign_convention:
        return np.conjugate(J)

    return J


def stokes_to_jones(S):
    """
    Convert a (list of) Stokes vector(s) to a (list of) Jones vector(s).

    The sign convention for the Jones vector can be set by calling
    `pypolar.jones.use_alternate_convention(True)`.  The default is to assume that
    the field is represented by exp(j*omega*t-k*z).

    Inputs:
        S : a single Stokes vector (4,) or list of Stokes vectors (n,4)

    Returns:
         a Jones vector (2,) or list of Jones vectors (n,2)
    """
    if S.ndim == 1:
        return _stokes_to_jones(S)

    n, m = S.shape
    if m != 4:
        print("Wrong shape ... should be %dx4 not %dx%d" % (m, n, m))
        return None

    J = np.empty(shape=(n, 2), dtype=np.ndarray)
    for i, SS in enumerate(S):
        J[i] = _stokes_to_jones(SS)

    return J


def mueller_to_jones(M):
    """
    Convert a Mueller matrix to a Jones matrix.

    Theocaris, Matrix Theory of Photoelasticity, eqns 4.70-4.76, 1979

    Inputs:
        M : a 4x4 Mueller matrix

    Returns:
         the corresponding 2x2 Jones matrix
    """
    A = np.empty((2, 2))
    A[0, 0] = np.sqrt((M[0, 0]+M[0, 1]+M[1, 0]+M[1, 1])/2)
    A[0, 1] = np.sqrt((M[0, 0]+M[0, 1]-M[1, 0]-M[1, 1])/2)
    A[1, 0] = np.sqrt((M[0, 0]-M[0, 1]+M[1, 0]-M[1, 1])/2)
    A[1, 1] = np.sqrt((M[0, 0]-M[0, 1]-M[1, 0]+M[1, 1])/2)

    theta = np.empty((2, 2))
    theta[0, 0] = 0
    theta[0, 1] = -np.arctan2(M[0, 3]+M[1, 3], M[0, 2]+M[1, 2])
    theta[1, 0] = np.arctan2(M[3, 0]+M[3, 1], M[2, 0]+M[2, 1])
    theta[1, 1] = np.arctan2(M[3, 2]-M[2, 3], M[2, 2]+M[3, 3])

    return A*np.exp(1j*theta)


def interpret(S):
    """
    Interpret a Stokes vector.

    Parameters
    S    : A Stokes vector

    Examples
    -------
    interpret([1, 0, 0, 0]) --> "Unpolarized Light"
    """
    try:
        S0, S1, S2, S3 = S
    except:
        print("Stokes vector must have four real elements")
        return 0

#    eps = 1e-12
    print("I = %.3f" % S0)
    print("Q = %.3f" % S1)
    print("U = %.3f" % S2)
    print("V = %.3f" % S3)

    s = "not implemented yet"
    return s



# pylint: disable=invalid-name
# pylint: disable=no-member

"""
Useful basic routines for managing symbolic Fresnel reflection.

To Do
    * tests and documentation

Scott Prahl
Apr 2019
"""

import sympy

__all__ = ('r_par_amplitude',
           'r_per_amplitude',
           't_par_amplitude',
           't_per_amplitude',
           'R_par',
           'R_per',
           'T_par',
           'T_per',
           'R_unpolarized',
           'T_unpolarized',
           'ellipsometry_rho',
           'ellipsometry_index')


def r_par_amplitude(m, theta_i):
    """
    Calculate the reflected amplitude for parallel polarized light.

    Args:
        m :       complex index of refraction   [-]
        theta_i : angle from normal to surface  [radians]
    Returns:
        reflected fraction of parallel field    [-]
    """
    c = m * m * sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s)
    if sympy.im(m) == 0:
        d = sympy.conjugate(d)
    rp = (c - d) / (c + d)
    return rp


def r_per_amplitude(m, theta_i):
    """
    Calculate the reflected amplitude for perpendicular polarized light.

    Args:
        m :       complex index of refraction     [-]
        theta_i : incidence angle from normal     [radians]
    Returns:
        reflected fraction of perpendicular field [-]
    """
    c = sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s)
    if sympy.im(m) == 0:
        d = sympy.conjugate(d)
    rs = (c - d) / (c + d)
    return rs


def t_par_amplitude(m, theta_i):
    """
    Calculate the transmitted amplitude for parallel polarized light.

    Args:
        m :       complex index of refraction  [-]
        theta_i : incidence angle from normal  [radians]
    Returns:
        transmitted fraction of parallel field [-]
    """
    c = sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s)
    if sympy.im(m) == 0:
        d = sympy.conjugate(d)
    tp = 2 * c * m / (m * m * c + d)
    return tp


def t_per_amplitude(m, theta_i):
    """
    Calculate the transmitted amplitude for perpendicular polarized light.

    Args:
        m :     complex index of refraction         [-]
        theta_i : incidence angle from normal       [radians]
    Returns:
        transmitted fraction of perpendicular field [-]
    """
    c = sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s)
    if sympy.im(m) == 0:
        d = sympy.conjugate(d)
    ts = 2 * c / (c + d)
    return ts


def R_par(m, theta_i):
    """
    Fraction of parallel-polarized light that is reflected (R_p).

    Calculate reflected fraction of incident power (or flux) assuming that
    the E-field of the incident light is parallel to the plane of incidence

    Args:
        m :       complex index of refraction [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected power                       [-]
    """
    return sympy.abs(r_par_amplitude(m, theta_i))**2


def R_per(m, theta_i):
    """
    Fraction of perpendicular-polarized light that is reflected (R_s).

    Calculate reflected fraction of incident power (or flux) assuming that
    the E-field of the incident light is perpendicular to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return sympy.abs(r_per_amplitude(m, theta_i))**2


def T_par(m, theta_i):
    """
    Fraction of parallel-polarized light that is transmitted (T_p).

    Calculate transmitted fraction of incident power (or flux) assuming that
    the E-field of the incident light is parallel to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        transmitted irradiance                [-]
    """
    c = sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s) # m*cos(theta_t)
    tp = 2 * c * m / (m * m * c + d)
    return d / c * sympy.abs(tp)**2


def T_per(m, theta_i):
    """
    Fraction of perpendicular-polarized light that is transmitted (T_s).

    Calculate transmitted fraction of incident power (or flux) assuming that
    the E-field of the incident light is perpendicular to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        transmitted field amplitude           [-]
    """
    c = sympy.cos(theta_i)
    s = sympy.sin(theta_i)
    d = sympy.sqrt(m * m - s * s) # m*cos(theta_t)
    ts = 2 * c / (c + d)
    return d / c * sympy.abs(ts)**2


def R_unpolarized(m, theta_i):
    """
    Fraction of unpolarized light that is reflected.

    Calculate reflection fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (R_par(m, theta_i) + R_per(m, theta_i)) / 2


def T_unpolarized(m, theta_i):
    """
    Fraction of unpolarized light that is transmitted.

    Calculate transmitted fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (T_par(m, theta_i) + T_per(m, theta_i)) / 2


def ellipsometry_rho(m, theta_i):
    """
    Calculate the ellipsometer parameter rho.

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        ellipsometer parameter rho            [-]
    """
    return r_par_amplitude(m, theta_i) / r_per_amplitude(m, theta_i)


def ellipsometry_index(rho, theta_i):
    """
    Calculate the index of refraction for an isotropic sample.

    Args:
        rho :  r_par_amplitude/r_per_amplitude                    [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        complex index of refraction           [-]
    """
    e_index = sympy.sqrt(1 - 4 * rho * sympy.sin(theta_i)**2 / (1 + rho)**2)
    return sympy.tan(theta_i) * e_index



import unittest
import numpy as np
import pypolar.jones as jones
import pypolar.mueller as mueller

class TestBasic(unittest.TestCase):

    def test_field_linear_H(self):
        H = jones.field_horizontal()
        J = jones.field_linear(0)
        for pair in zip(H, J):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_field_linear_V(self):
        V = jones.field_vertical()
        J = jones.field_linear(np.pi/2)
        for pair in zip(V, J):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_field_linear_multi(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        H = jones.field_horizontal()
        V = jones.field_vertical()
        J = jones.field_linear(angles)
        self.assertEqual(len(J),N)

    def test_intensity_scalar(self):
        J = jones.field_right_circular()
        I = jones.intensity(J)
        self.assertAlmostEqual(I,1)
        J = jones.field_horizontal()
        I = jones.intensity(J)
        self.assertAlmostEqual(I,1)

    def test_intensity_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        J = jones.field_linear(angles)
        I = jones.intensity(J)
        self.assertEqual(len(I),N)
        for intensity in I:
            self.assertAlmostEqual(intensity, 1)

    def test_phase_scalar(self):
        J = jones.field_left_circular()
        phi = jones.phase(J)
        self.assertAlmostEqual(phi, -np.pi/2)
        J = jones.field_right_circular()
        phi = jones.phase(J)
        self.assertAlmostEqual(phi, np.pi/2)
        J = jones.field_horizontal()
        phi = jones.phase(J)
        self.assertEqual(phi,0)

    def test_phase_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        J = jones.field_linear(angles)
        phi = jones.phase(J)
        self.assertEqual(len(phi),N)
        for p in phi:
            self.assertAlmostEqual(p, 0)

    def test_to_stokes_scalar(self):
        J = jones.field_left_circular()
        S = mueller.stokes_left_circular()
        SS = jones.jones_to_stokes(J)
        for pair in zip(S, SS):
            self.assertAlmostEqual(pair[0], pair[1])

    def test_to_stokes_array(self):
        N = 3
        angles = np.linspace(0,np.pi/2,N)
        J = jones.field_linear(angles)
        S = jones.jones_to_stokes(J)
        n,m = S.shape
        self.assertEqual(n,N)
        self.assertEqual(m,4)

if __name__ == '__main__':
    unittest.main()


# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
"""
Useful functions for ellipsometry.

Scott Prahl
Apr 2020
"""

import numpy as np
import pypolar.fresnel

__all__ = ('rho_from_m',
           'rho_from_tanpsi_Delta',
           'tanpsi_Delta_from_rho',
           'm_from_rho',
           'm_from_tanpsi_and_Delta',
           'rho_from_zone_2_null_angles',
           'rho_from_zone_4_null_angles',
           'rho_from_rotating_analyzer_data',
           'null_angles',
           'null_angles_report',
           'rotating_analyzer_signal',
           'RAE_from_rho',
           'rotating_analyzer_signal_from_rho',
           'rotating_analyzer_signal_from_m',
           'find_fourier',
           'rho_from_rotating_analyzer_data',
           'rho_from_PSA',
           'm_from_rotating_analyzer_data',
           )


def rho_from_m(m, theta_i):
    """
    Calculate the complex ratio of reflection amplitudes.

    This assumes that the material is flat and isotropic (e.g., no
    surface film).  It also assumes that the parallel (or perpendicular)
    field remains entirely parallel (or perpendicular) and is fully
    characterized by Fresnel reflection.

    Args:
        m:       complex index of refraction   [-]
        theta_i: incidence angle from normal [radians]
    Returns:
        complex ellipsometer parameter rho    [-]
    """
    rp = pypolar.fresnel.r_par_amplitude(m, theta_i)
    rs = pypolar.fresnel.r_per_amplitude(m, theta_i)
    return rp/rs


def rho_from_tanpsi_Delta(tanpsi, Delta):
    """
    Calculate the index of refraction for an isotropic sample.

    Formula from McCrackin "Measurement of the thickness and refractive
    index of very thin films and the optical properties of surfaces by
    ellipsometry", Journal of Research of the National Bureau of Standards,
    (1963).

    Args:
        tanpsi : tan(psi) or abs(rpar/rperp)               [-]
        Delta :  phase change caused by reflection         [radians]
    Returns:
        complex ellipsometer parameter rho                 [-]
    """
    return tanpsi*np.exp(1j*Delta)


def tanpsi_Delta_from_rho(rho):
    """
    Extract ellipsometer parameters from rho.

    Formula from Fujiwara 2007 eqn 4.6 and correspond to the case
    when the complex refractive index is negative (m = n-k*1j)

    Args:
        rho :  r_par_amplitude/r_per_amplitude or tan(psi)*exp(j*Delta)  [-]
    Returns:
        tanpsi : tan(psi) or abs(r_p/r_s)                  [-]
        Delta :  phase change caused by reflection         [radians]
    """
    Delta = np.arctan2(rho.imag, rho.real)
    if rho.real < 0:
        if rho.imag > 0:
            Delta += np.pi
        else:
            Delta -= np.pi
    tanpsi = abs(rho)
    return tanpsi, Delta


def m_from_rho(rho, theta_i):
    """
    Calculate the index of refraction for an isotropic sample.

    Formula from McCrackin "Measurement of the thickness and refractive
    index of very thin films and the optical properties of surfaces by
    ellipsometry", Journal of Research of the National Bureau of Standards,
    (1963).

    Args:
        rho :  r_par_amplitude/r_per_amplitude or tan(psi)*exp(j*Delta)        [-]
        theta_i : incidence angle from normal              [radians]
    Returns:
        complex index of refraction                        [-]
    """
    e_index = np.sqrt(1 - 4 * rho * np.sin(theta_i)**2 / (1 + rho)**2)

    # choose proper branch
    if np.isscalar(rho):
        if np.angle(rho) < 0:
            e_index = np.conjugate(e_index)
    else:
        for i, r in enumerate(rho):
            if np.angle(r) < 0:
                e_index[i] = np.conjugate(e_index[i])

    return np.tan(theta_i) * e_index


def m_from_tanpsi_and_Delta(tanpsi, Delta, theta_i):
    """
    Return the index of refraction for observed Delta, tanpsi, and theta_i.

    Args:
        tanpsi : abs() of ratio of field amplitudes        [-]
        Delta :  phase change caused by reflection         [-]
        theta_i : incidence angle from normal              [radians]
    Returns:
        complex index of refraction                        [-]
    """
    rho = rho_from_tanpsi_Delta(tanpsi, Delta)
    return m_from_rho(rho, theta_i)


def rho_from_zone_2_null_angles(P2, A2):
    """
    Recover rho from Null ellipsometer measurements in zone 2.

    Args:
        P2: polarizer angle for null reading     [radians]
        A2: analyzer angle for null reading     [radians]
    Returns:
        complex ellipsometer parameter rho                 [-]
    """
    if (A2 < 0 or A2 > np.pi/2):
        print("Analyzer is not zone 2 (0 < %.2f < pi/2)" % P2)
        return 0
    if (P2 < -np.pi/4 or P2 > 3*np.pi/4):
        print("Polarizer is not zone 2 (-pi/4 < %.2f < 3pi/4)" % A2)
        return 0
    psi = A2
    Delta = 3*np.pi/2 - 2*P2
    rho = rho_from_tanpsi_Delta(np.tan(psi), Delta)
    return rho


def rho_from_zone_4_null_angles(P4, A4):
    """
    Recover rho from Null ellipsometer measurements in zone 4.

    Args:
        P4: polarizer angle for null reading in zone 4    [radians]
        A4: analyzer angle for null reading in zone 4   [radians]
    Returns:
        complex ellipsometer parameter rho                 [-]
    """
    if (A4 < -np.pi/2 or A4 > 0):
        print("Analyzer is not zone 4 (-pi/2 < %.2f < 0)" % P4)
        return 0
    if (P4 < -3*np.pi/4 or P4 > np.pi/4):
        print("Polarizer is not zone 4 (-3pi/4 < %.2f < pi/4)" % A4)
        return 0
    psi = -A4
    Delta = np.pi/2 - 2*P4
    rho = rho_from_tanpsi_Delta(np.tan(psi), Delta)
    return rho


def null_angles(m, theta_i):
    """
    Generate expected ellipsometer angles for all zones.

    The various null angles fall into four sets called zones, two with the fast-axis
    of the quarter wave plate set 45° (2 & 4) and two with the fast-axis of the
    quarter wave plate set to -45° (1 & 3).  In each zone there are four combinations
    of polarizer and analyzer angles that have a null reading (because rotation of a
    linear polarizer by 180° should give the same result.

    Table 1 from McCrackin "Measurement of the thickness and refractive
    index of very thin films and the optical properties of surfaces by
    ellipsometry", Journal of Research of the National Bureau of Standards,
    (1963).

    All angles returned fall between 0 and 2pi.

    Args:
        m:       complex index of refraction   [-]
        theta_i: incidence angle from normal [radians]
    Returns:
        dictionary with null angles [(P1, A1), (P2, A2), (P3, A3), (P4, A4)] for each zone
    """
    rho = rho_from_m(m, theta_i)
    tanpsi = np.abs(rho)
    psi = np.arctan(tanpsi)
    Delta = np.angle(rho)
    p = Delta/2-np.pi/4
    a = psi
    pi = np.pi
    PA = {}
    PA[1] = np.array([(p, a), (p+pi, a), (p, a+pi), (p+np.pi, np.pi)])
    PA[2] = np.array([(pi/2-p, a), (3*pi/2-p, a), (pi/2-p, a+pi), (3*pi/2-p, a+np.pi)])
    PA[3] = np.array([(p+pi/2, pi-a), (p+3*pi/2, pi-a), (p+pi/2, 2*pi-a), (p+3*pi/2, 2*pi-a)])
    PA[4] = np.array([(pi-p, pi-a), (2*pi-p, pi-a), (pi-p, 2*pi-a), (2*pi-p, 2*pi-a)])

    # make all measurements between 0 and 2pi
    PA[1] = np.remainder(PA[1] +2*np.pi, 2*np.pi)
    PA[2] = np.remainder(PA[2] +2*np.pi, 2*np.pi)
    PA[3] = np.remainder(PA[3] +2*np.pi, 2*np.pi)
    PA[4] = np.remainder(PA[4] +2*np.pi, 2*np.pi)

    return PA


def null_angles_report(m, theta_i):
    """
    Create a report showing null angles for sample.

    Args:
        m:       complex index of refraction   [-]
        theta_i: incidence angle from normal [radians]
    Returns:
        string containing a report listing null angles for each zone.
    """
    pa = null_angles(m, theta_i)

    rho = rho_from_m(m, theta_i)
    tanpsi = np.abs(rho)
    psi = np.arctan(tanpsi)
    Delta = np.angle(rho)
    p = Delta/2-np.pi/4
    a = psi

    s = "m       = %.4f%+.4fj\n" % (m.real, m.imag)
    s += "theta_i = %7.1f°\n" % np.degrees(theta_i)
    s += '\n'

    s += "zone  P   theta_a\n"
    for zone in [1, 3, 2, 4]:
        for pair in pa[zone]:
            thetap, thetaa = np.degrees(pair)
            s += "  %d  %7.1f°  %7.1f°\n" % (zone, thetap, thetaa)
        s += '\n'

    s += "p       = %7.1f°\n" % np.degrees(p)
    s += "a       = %7.1f°\n" % np.degrees(a)
    s += '\n'
    s += "psi     = %7.1f°\n" % np.degrees(psi)
    s += "Delta   = %7.1f°\n" % np.degrees(Delta)
    s += '\n'

    return s


def rotating_analyzer_signal(phi, IDC, IS, IC, noise=0):
    """
    Create theoretical rotating ellipsometer signal.

    In theory the rotating analyzer ellipsometer generates a
    sinusoidal signal with an offset.  This function does that
    and allows the optional addition of normally distributed
    noise.

    Args:
        phi    array of analyzer angles                [radians]
        IDC:   DC amplitude of signal                     [-]
        IS:    sin(2*phi) amplitude coefficient           [-]
        IC:    cos(2*phi) amplitude coefficient           [-]
        noise: std dev of normal noise distribution       [-]
    Returns:
        Array of ellipsometer readings for each angle phi [-]
    """
    base = IDC+IS*np.sin(2*phi)+IC*np.cos(2*phi)
    noise = np.random.normal(0, noise, len(phi))
    return base+noise


def rotating_analyzer_signal_from_rho_old(phi, rho, P, QWP=False, average=1, noise=0):
    """
    Create normalized rotating ellipsometer signal for sample.

    Generate the expected reading at each analyzer angle in an ellipsometer
    with a sample characterized by a material with an ellipsometer parameter
    rho = tan(psi)exp(j*Delta)

    This is a classic
    source :: polarizer :: QWP :: sample :: analyzer :: detector
    arrangement.  The QWP is oriented at +45° if present.

    Note that the default returned array is normalized between 0 and 1.
    therefore the noise should be scaled accordingly.

    Args:
        phi:     array of analyzer angles from 0 to 2pi   [radians]
        rho:     ellipsometer parameter for surface       [complex]
        P:       angle of polarizer                       [radians]
        QWP:     True if QWP is present
        average: average value of signal over 2pi         [AU]
        noise:   std dev of normal noise distribution     [AU]
    Returns:
        Array of ellipsometer readings for each angle phi [-]
    """
    tanpsi = np.abs(rho)
    Delta = np.angle(rho)
    if QWP:
        Delta -= np.pi/2
    denom = tanpsi**2+np.tan(P)**2
    alpha = (tanpsi**2-np.tan(P)**2)/denom
    beta = 2*tanpsi*np.cos(Delta)*np.tan(P)/denom
    base = 1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi)
    noise = np.random.normal(0, noise, len(phi))
    return average*base+noise


def RAE_from_rho(phi, rho, P, average=1, noise=0):
    """
    Create normalized rotating ellipsometer signal for sample.

    See eqn 4.19 and eqn 4.23 in Fujiwara 2007

    Generate the expected reading at each analyzer angle in an ellipsometer
    with a sample characterized by a material with an ellipsometer parameter
    rho = tan(psi)exp(j*Delta)

    This is a classic
    source :: polarizer :: QWP :: sample :: analyzer :: detector
    arrangement.  The QWP is oriented at +45° if present.

    Note that the default returned array is normalized between 0 and 1.
    therefore the noise should be scaled accordingly.

    Args:
        phi:     array of analyzer angles from 0 to 2pi   [radians]
        rho:     ellipsometer parameter for surface       [complex]
        P:       angle of polarizer                       [radians]
        average: average value of signal over 2pi         [AU]
        noise:   std dev of normal noise distribution     [AU]
    Returns:
        Array of ellipsometer readings for each angle phi [-]
    """
    tanpsi = np.abs(rho)
    Delta = np.angle(rho)
    tanP = np.tan(P)
    denom = tanpsi**2 + tanP**2
    alpha = (tanpsi**2 - tanP**2)/denom
    beta = (2*tanpsi*np.cos(Delta)*tanP)/denom
    base = 1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi)

    noise = np.random.normal(0, noise, len(phi))
    return average*base+noise


def rotating_analyzer_signal_from_rho(phi, rho, P, QWP=False, average=1, noise=0):
    """
    Create normalized rotating ellipsometer signal for sample.

    Generate the expected reading at each analyzer angle in an ellipsometer
    with a sample characterized by a material with an ellipsometer parameter
    rho = tan(psi)exp(j*Delta)

    This is a classic
    source :: polarizer :: QWP :: sample :: analyzer :: detector
    arrangement.  The QWP is oriented at +45° if present.

    Note that the default returned array is normalized between 0 and 1.
    therefore the noise should be scaled accordingly.

    Args:
        phi:     array of analyzer angles from 0 to 2pi   [radians]
        rho:     ellipsometer parameter for surface       [complex]
        P:       angle of polarizer                       [radians]
        QWP:     True if QWP is present
        average: average value of signal over 2pi         [AU]
        noise:   std dev of normal noise distribution     [AU]
    Returns:
        Array of ellipsometer readings for each angle phi [-]
    """
    tanpsi = np.abs(rho)
    Delta = np.angle(rho)
    if QWP:
        psi = np.arctan(tanpsi)
        alpha = -np.cos(2*psi)
        beta = np.sin(2*psi)*np.cos(Delta + 2 * P)
    else:
        tanP = np.tan(P)
        denom = tanpsi**2 + tanP**2
        alpha = (tanpsi**2 - tanP**2)/denom
        beta = (2*tanpsi*np.cos(Delta)*tanP)/denom
    base = 1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi)
    noise = np.random.normal(0, noise, len(phi))
    return average*base+noise


def rotating_analyzer_signal_from_m(phi, m, theta_i, P, average=1, noise=0):
    """
    Create rotating ellipsometer signal for sample with known index.

    Args:
        phi:     array of analyzer angles from 0 to 2pi   [radians]
        m:       complex index of refraction of sample    [-]
        theta_i: angle of incidence (from normal)         [radians]
        P:       angle of incident polarized light        [radians]
        average: average value of signal over 2pi         [AU]
        noise:   std dev of normal noise distribution     [-]

    Returns:
        Array of ellipsometer readings for each angle phi [-]

    """
    sig = pypolar.fresnel.r_par_amplitude(m, theta_i)*np.cos(P)*np.cos(phi)
    sig += pypolar.fresnel.r_per_amplitude(m, theta_i)*np.sin(P)*np.sin(phi)
    base = np.cos(P)**2*abs(sig)**2
    noise = np.random.normal(0, noise, len(phi))
    return average*base + noise


def find_fourier(phi, signal):
    """
    Calculate first few Fourier series coefficients.

    Fit the signal to the function
        I_ave * ( 1 + alpha*cos(2*phi) + beta*sin(2*phi) )

    args:
        phi:    array of analyzer angles
        signal: array of ellipsometer intensities
    returns:
        I_ave, alpha, beta
    """
    I_ave = np.average(signal)
    I_S = 2*np.average(signal*np.sin(2*phi))
    I_C = 2*np.average(signal*np.cos(2*phi))
    alpha = I_C / I_ave
    beta = I_S / I_ave
    return I_ave, alpha, beta


def rho_from_rotating_analyzer_data_old(phi, signal, P, QWP=False):
    """
    Recover rho from rotating analyzer data (old version).

    This is done by fitting the signal to
             I_ave * (1 + alpha*cos(2*phi) + beta*sin(2*phi))
    Then alpha and beta are used to find tan(psi) and Delta following
    e.g. eqn 4.24 in Fujiwara.

    Args:
        phi:     array of analyzer angles               [radians]
        signal:  array of ellipsometer intensities      [AU]
        P:       incident polarization azimuthal angle  [radians]
        QWP:     True if QWP is present
    Returns:
        rho = tan(psi)*exp(1j*Delta)                    [-]
        fit: array of fitted data
    """
    I_ave, alpha, beta = find_fourier(phi, signal)

    tanP = np.tan(P)
    arg = beta / np.sqrt(abs(1 - alpha**2)) * np.sign(tanP)
    if arg > 1:
        Delta = 0
    elif arg < -1:
        Delta = np.pi
    else:
        Delta = np.arccos(arg)

    if QWP:
        Delta += np.pi/2

    tanpsi = np.sqrt(abs(1 + alpha) / abs(1 - alpha)) * np.abs(tanP)
    rho = tanpsi * np.exp(1j*Delta)

    fit = I_ave * (1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi))
    return rho, fit


def rho_from_rotating_analyzer_data(phi, signal, P, QWP=False):
    """
    Recover rho from rotating analyzer data.

    Based on equation 3.297 from Azzam
    (should be fixed to work with any P value)

    Args:
        phi:     array of analyzer angles               [radians]
        signal:  array of ellipsometer intensities      [AU]
        P:       incident polarization azimuthal angle  [radians]
        QWP:     True if QWP is present
    Returns:
        rho = tan(psi)*exp(1j*Delta)                    [-]
        fit: array of fitted data
    """
    I_ave, alpha, beta = find_fourier(phi, signal)

    if QWP:
        tanPC = np.tan(P+np.pi/4)
        factor = (1+1j*tanPC)/(1-1j*tanPC)
    else:
        factor = np.tan(P)

    delta = complex(1-alpha**2-beta**2, 0)
    rho = (1+alpha)/(beta-1j*np.sqrt(delta)) * factor
    rho *= np.exp(-1j*np.pi/2*QWP)

    fit = I_ave * (1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi))

    if 0 <= P <= np.pi/2:
        return rho, fit

    return  np.conjugate(rho), fit


def rho_from_PSA(phi, signal, P):
    """
    Recover rho from polarizer/sample/rotating analyzer system.

    Based on equation 4.24 in Fujiwara 2005.  Note that the PSA
    arrangement allows 0<=psi<=90° and 0<=Delta<=180°.

    In this system the measurement error increases when Delta is
    near zero or 180°.  Since this corresponds to linearly polarized
    light and would be exactly what would be measured for dielectric
    samples.

    Args:
        phi:     array of analyzer angles               [radians]
        signal:  array of ellipsometer intensities      [AU]
        P:       incident polarization azimuthal angle  [radians]
    Returns:
        rho = tan(psi)*exp(1j*Delta)                    [-]
    """
    I_ave, alpha, beta = find_fourier(phi, signal)
    tanpsi = np.sqrt((1+alpha)/(1-alpha)) * abs(np.tan(P))
    Delta = np.arccos(beta/np.sqrt(1-alpha**2))
    rho = tanpsi * np.exp(1j*Delta)
    fit = I_ave * (1 + alpha * np.cos(2*phi) + beta * np.sin(2*phi))
    return rho, fit


def m_from_rotating_analyzer_data(phi, signal, theta_i, P, QWP=False):
    """
    Recover m from rotating analyzer data.

    Args:
        phi:     array of analyzer angles               [radians]
        signal:  array of ellipsometer intensities      [AU]
        theta_i: incidence angle from normal            [radians]
        P:       incident polarization azimuthal angle  [radians]
        QWP:     True if QWP is present
    Returns:
        complex index of refraction                     [-]
    """
    rho, fit = rho_from_rotating_analyzer_data(phi, signal, P, QWP)
    m = m_from_rho(rho, theta_i)
    return m, fit



# pylint: disable=invalid-name
"""
Simple interface to Gaerter 116A Ellipsometer.

This is intended to be used with the computer connected to an
Arduino controller that interfaces with the Gaerter ellipsometer.

Scott Prahl
Apr 2020
"""
import os
import time
import serial
import serial.tools.list_ports
import numpy as np

__all__ = ('get_reflectance',
           'avg_reflectance',
           'current_serial_ports',
           'connect_to_ellipsometer',
           'save_data_with_time_stamp',
           'save_data_with_name',
           'read_data_with_name',
           )


def current_serial_ports():
    """Return a string listing all current serial port names."""
    ports = list(serial.tools.list_ports.comports())
    s = "%35s  %s\n" % ('portname', 'info')
    for p in ports:
        s += "%35s  %s\n" %(p[0], p[1])
    return s


def connect_to_ellipsometer(usb_serial_port_id):
    """
    Establish a serial connection with the ellipsometer.

    Args:
        usb_serial_port_id: name of serial port
    Returns
        Serial connection object
    """
    # 38400 is the fastest baud rate that works
    baud_rate = 38400

    x = os.system("ls " + usb_serial_port_id)

    if x != 0:
        print("ERROR: serial port '%s' is disconnected" % usb_serial_port_id)
        return None

    try:
        conn = serial.Serial(usb_serial_port_id, baud_rate, timeout=2)

    except serial.serialutil.SerialException:
        print('Error: No serial connection to Arduino established!')
        print('       Correct the serial port id above')
        print('       Use list_serial_ports() to find possible names')
        return None

    time.sleep(2)                  # wait for connection to stablize
    return conn


def get_reflectance(conn):
    """
    Obtain one analyzer revolution worth of reflectance values.

    Trigger the Arduino attached to the ellipsometer and obtain
    readings at each of 72 analyzer orientations.

    The idea is pretty simple. Just send a byte to the Arduino and
    it will send 144=72*2 bytes back.
    returns:
        an array of 72 integers
    """
    N = 72
    conn.write(b'1')               # trigger the Arduino
    a = bytearray(conn.read(2*N))  # read 144 bytes

    if len(a) != 144:
        print("Bad number of bytes read (%d != 144)" % len(a))
        print("Make sure ellipsometer is rotating!")
        return np.zeros(72, dtype=np.int64)

    # assume Arduino and host have same endianness.  Reassemble as 16 ints
    y = [(a[2*i] << 8) + a[2*i+1] for i in range(N)]
    return np.array(y)


def avg_reflectance(conn, num_samples=10):
    """
    Collect multiple samples and return the sum of all of them.

    returns:
        an array of 72 integers
    """
    total = get_reflectance(conn)

    for _i in range(1, num_samples):
        total += get_reflectance(conn)

    return total // num_samples


def save_data_with_time_stamp(signal):
    """
    Save signal in time-stamped file.

    Creates a filename using the current time.

    Args:
        signal: array of ellipsometer readings
    Returns:
        filename of the file created
    """
    t = time.localtime()
    time_stamp_name = time.strftime('lab3-%Y-%m-%d-%H-%M-%S.txt', t)
    np.savetxt(time_stamp_name, signal, fmt="%0.f", newline=', ')
    print("saved data to file named '%s'" % time_stamp_name)
    return time_stamp_name


def save_data_with_name(signal, basename, theta_i, P, QWP=False):
    """
    Save signal to file with ellipsometer setting in the filename.

    Creates a filename based on the settings and saves the array signal to it.

    Args:
        signal:   array of ellipsometer readings
        basename: string starting filename
        theta_i:  angle of incidence               [radians]
        P:        polarizer angle                  [radians]
        QWP:      True if QWP is present
    Returns:
        filename of the file created
    """
    I = np.degrees(theta_i)
    PP = np.degrees(P)

    t = time.localtime()
    time_stamp_name = time.strftime('%Y-%m-%d-%H-%M-%S.txt', t)

    filename = "%s, I=%.1f°, P=%.1f°, QWP=%r, " % (basename, I, PP, QWP)
    filename = filename + time_stamp_name
    np.savetxt(filename, signal, fmt="%0.f", newline=', ')

    print("saved data to file named '%s'" % filename)
    return filename


def read_data_with_name(filename):
    """
    Read a text file into an array.

    This does the opposite of `save_data_with_time_stamp()` or `save_data_with_name()`
    The data is read and returned in a 72 element array.
    """
    x = os.system("ls " + filename)

    if x != 0:
        print("ERROR: file '%s' does not exist" % filename)
        return None

    signal = np.genfromtxt(filename, delimiter=', ')
    signal = np.delete(signal, -1, 0) # drop the last NaN value
    return signal
